# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable
from catalystwan.models.configuration.feature_profile.common import Prefix

NetworkType = Literal[
    "broadcast",
    "point-to-point",
    "non-broadcast",
    "point-to-multipoint",
]

NoAuthType = Literal["no-auth"]
IpsecSha1AuthType = Literal["ipsec-sha1"]

MaxMetricRouterLsaAction = Literal[
    "disabled",
    "immediately",
    "on-startup",
]

RedistributeProtocol = Literal[
    "static",
    "connected",
    "bgp",
    "omp",
    "nat-route",
    "eigrp",
]

RedistributeProtocolIPv6 = Literal[
    "static",
    "connected",
    "bgp",
    "omp",
    "eigrp",
]
MetricType = Literal["type1", "type2"]


class NoAuth(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    auth_type: Union[Global[NoAuthType], Default[NoAuthType]] = Field(
        serialization_alias="authType",
        validation_alias="authType",
        default=Default[NoAuthType](value="no-auth"),
    )


class IpsecSha1Auth(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    auth_type: Global[IpsecSha1AuthType] = Field(
        serialization_alias="authType",
        validation_alias="authType",
        default=Global[IpsecSha1AuthType](value="ipsec-sha1"),
    )
    spi: Union[Global[int], Variable]
    auth_key: Union[Global[str], Variable] = Field(serialization_alias="authKey", validation_alias="authKey")


class Ospfv3InterfaceParametres(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: Optional[Union[Global[str], Variable]] = Field(serialization_alias="ifName", validation_alias="ifName")
    hello_interval: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="helloInterval", validation_alias="helloInterval", default=None
    )
    dead_interval: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="deadInterval", validation_alias="deadInterval", default=None
    )
    retransmit_interval: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="retransmitInterval", validation_alias="retransmitInterval", default=None
    )
    cost: Optional[Union[Global[int], Variable, Default[None]]] = None
    priority: Optional[Union[Global[int], Variable, Default[int]]] = None
    network_type: Optional[Union[Global[NetworkType], Variable, Default[None]]] = Field(
        serialization_alias="networkType", validation_alias="networkType", default=None
    )
    passive_interface: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="passiveInterface", validation_alias="passiveInterface", default=None
    )
    authentication_config: Optional[Union[NoAuth, IpsecSha1Auth]] = Field(
        serialization_alias="authenticationConfig", validation_alias="authenticationConfig", default=None
    )


class SummaryRouteIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    network: Union[Global[str], Variable]
    no_advertise: Union[Global[bool], Variable, Default[bool]] = Field(
        serialization_alias="noAdvertise", validation_alias="noAdvertise"
    )
    cost: Optional[Union[Global[int], Variable, Default[None]]] = None


class SummaryRoute(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    network: Optional[Prefix] = None
    cost: Optional[Union[Global[int], Variable, Default[None]]] = None
    no_advertise: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="noAdvertise", validation_alias="noAdvertise", default=None
    )


class StubArea(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    area_type: Global[str] = Field(
        serialization_alias="areaType", validation_alias="areaType", default=Global[str](value="stub")
    )
    no_summary: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="noSummary", validation_alias="noSummary", default=None
    )


class NssaArea(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    area_type: Global[str] = Field(
        serialization_alias="areaType", validation_alias="areaType", default=Global[str](value="nssa")
    )
    no_summary: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="noSummary", validation_alias="noSummary", default=None
    )
    always_translate: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="alwaysTranslate", validation_alias="alwaysTranslate", default=None
    )


class NormalArea(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    area_type: Global[str] = Field(
        serialization_alias="areaType", validation_alias="areaType", default=Global[str](value="normal")
    )


class DefaultArea(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    area_type: Default[None] = Field(
        serialization_alias="areaType", validation_alias="areaType", default=Default[None](value=None)
    )


class Ospfv3IPv4Area(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    area_number: Union[Global[int], Variable] = Field(serialization_alias="areaNum", validation_alias="areaNum")
    area_type_config: Optional[Union[StubArea, NssaArea, NormalArea, DefaultArea]] = Field(
        serialization_alias="areaTypeConfig", validation_alias="areaTypeConfig", default=None
    )
    interfaces: List[Ospfv3InterfaceParametres]
    ranges: Optional[List[SummaryRoute]] = None


class Ospfv3IPv6Area(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    area_number: Union[Global[int], Variable] = Field(serialization_alias="areaNum", validation_alias="areaNum")
    area_type_config: Optional[Union[StubArea, NssaArea, NormalArea, DefaultArea]] = Field(
        serialization_alias="areaTypeConfig", validation_alias="areaTypeConfig", default=None
    )
    interfaces: List[Ospfv3InterfaceParametres]
    ranges: Optional[List[SummaryRoute]] = None


class MaxMetricRouterLsa(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    action: Global[MaxMetricRouterLsaAction]
    on_startup_time: Optional[Union[Global[int], Variable]] = Field(
        serialization_alias="onStartUpTime", validation_alias="onStartUpTime", default=None
    )


class RedistributedRoute(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    protocol: Union[Global[RedistributeProtocol], Variable]
    nat_dia: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="natDia", validation_alias="natDia", default=None
    )
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy", default=None
    )


class RedistributedRouteIPv6(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    protocol: Union[Global[RedistributeProtocolIPv6], Variable]
    route_policy: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="routePolicy", validation_alias="routePolicy", default=None
    )


class DefaultOriginate(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    originate: Union[Global[bool], Default[bool]]
    always: Optional[Union[Global[bool], Variable, Default[bool]]] = None
    metric: Optional[Union[Global[str], Variable, Default[None]]] = None
    metricType: Optional[Union[Global[MetricType], Variable, Default[None]]] = None


class SpfTimers(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    delay: Optional[Union[Global[int], Variable, Default[int]]] = None
    initial_hold: Optional[Union[Global[int], Variable, Default[int]]] = None
    max_hold: Optional[Union[Global[int], Variable, Default[int]]] = None


class AdvancedOspfv3Attributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    reference_bandwidth: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="referenceBandwidth", validation_alias="referenceBandwidth", default=None
    )
    compatible_rfc1583: Optional[Union[Global[bool], Variable, Default[bool]]] = Field(
        serialization_alias="compatibleRfc1583", validation_alias="compatibleRfc1583", default=None
    )
    default_originate: Optional[DefaultOriginate] = Field(
        serialization_alias="defaultOriginate", validation_alias="defaultOriginate", default=None
    )
    spf_timers: Optional[SpfTimers] = Field(serialization_alias="spfTimers", validation_alias="spfTimers", default=None)
    policy_name: Optional[Union[Default[None], Global[UUID]]] = Field(
        serialization_alias="policyName", validation_alias="policyName", default=None
    )
    filter: Optional[Union[Global[bool], Variable, Default[bool]]] = None


class BasicOspfv3Attributes(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    router_id: Optional[Union[Global[str], Variable, Default[None]]] = Field(
        serialization_alias="routerId", validation_alias="routerId", default=None
    )
    distance: Optional[Union[Global[int], Variable, Default[int]]] = None
    external_distance: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="externalDistance", validation_alias="externalDistance", default=None
    )
    inter_area_distance: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="interAreaDistance", validation_alias="interAreaDistance", default=None
    )
    intra_area_distance: Optional[Union[Global[int], Variable, Default[int]]] = Field(
        serialization_alias="intraAreaDistance", validation_alias="intraAreaDistance", default=None
    )


class Ospfv3IPv4Data(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    basic: Optional[BasicOspfv3Attributes] = None
    advanced: Optional[AdvancedOspfv3Attributes] = None
    redistribute: Optional[RedistributedRoute] = None
    max_metric_router_lsa: Optional[MaxMetricRouterLsa] = Field(
        serialization_alias="maxMetricRouterLsa", validation_alias="maxMetricRouterLsa", default=None
    )
    area: List[Ospfv3IPv4Area]


class Ospfv3IPv6Data(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    basic: Optional[BasicOspfv3Attributes] = None
    advanced: Optional[AdvancedOspfv3Attributes] = None
    redistribute: Optional[RedistributedRouteIPv6] = None
    max_metric_router_lsa: Optional[MaxMetricRouterLsa] = Field(
        serialization_alias="maxMetricRouterLsa", validation_alias="maxMetricRouterLsa", default=None
    )
    area: List[Ospfv3IPv6Area]


class Ospfv3IPv4CreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: Ospfv3IPv4Data


class Ospfv3IPv6CreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: Ospfv3IPv6Data
    metadata: Optional[dict] = None



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar, List, Optional

from attr import define, field  # type: ignore
from pydantic import ConfigDict, field_validator

from catalystwan.api.templates.feature_template import FeatureTemplate
from catalystwan.api.templates.payloads.aaa.aaa_model import VpnType

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class GatewayType(Enum):
    NEXT_HOP = "next-hop"
    NULL_0 = "null0"
    VPN = "vpn"
    DHCP = "dhcp"


@define
class Mapping:
    name: str
    ips: List[str] = field(factory=list)


@define
class DNS:
    primary: str
    secondary: Optional[str] = None
    primaryv6: Optional[str] = None
    secondaryv6: Optional[str] = None


@define
class NextHop:
    address: str
    distance: int = 1
    en_distance: bool = False


@define
class IPv4Route:
    prefix: str
    next_hop: list[NextHop]
    gateway: GatewayType = field(default=GatewayType.NEXT_HOP)


@define
class IPv6Route:
    prefixv6: str
    next_hopv6: list[NextHop]
    gatewayv6: GatewayType = field(default=GatewayType.NEXT_HOP)


class CiscoVPNModel(FeatureTemplate):
    type: ClassVar[str] = "cisco_vpn"  # Cisco VPN
    payload_path: ClassVar[Path] = Path(__file__).parent / "feature/cisco_vpn.json.j2"
    tenant_vpn: Optional[int] = None
    tenant_org_name: Optional[str] = None
    vpn_id: int
    dns: Optional[DNS] = None
    mapping: List[Mapping] = []
    ipv4route: List[IPv4Route] = []
    ipv6route: List[IPv6Route] = []

    @field_validator("vpn_id")
    @classmethod
    def check_id(cls, v: int):
        if v not in [VpnType.VPN_TRANSPORT.value, VpnType.VPN_MANAGMENT.value]:
            if cls.tenant_org_name is None:
                raise ValueError("Must enter the name of the organization.")
        return v

    def generate_vpn_id(self, session: ManagerSession) -> None:
        self.tenant_vpn = self.vpn_id
        payload = {"resourcePoolDataType": "vpn", "tenantId": self.tenant_org_name, "tenantVpn": self.tenant_vpn}
        url = "/dataservice/resourcepool/resource/vpn"
        response = session.put(url=url, json=payload).json()
        self.vpn_id = response["deviceVpn"]

    def generate_payload(self, session: ManagerSession) -> str:
        if self.vpn_id not in [0, 512]:
            self.generate_vpn_id(session=session)
        return super().generate_payload(session)

    model_config = ConfigDict(arbitrary_types_allowed=True)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from enum import Enum


class Severity(Enum):
    MINOR = "Minor"
    MEDIUM = "Medium"
    MAJOR = "Major"
    CRITICAL = "Critical"
    UNKNOWN = None


class Viewed(Enum):
    YES = "true"
    NO = "false"



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import unittest
from unittest.mock import patch
from uuid import uuid4

from parameterized import parameterized  # type: ignore

from catalystwan.api.mtt_aaa_api import TenantAaaAPI, TenantRadiusAPI, TenantTacacsAPI
from catalystwan.dataclasses import TenantAAA, TenantRadiusServer, TenantTacacsServer
from catalystwan.utils.creation_tools import create_dataclass


class TestAaaAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.aaa = {
            "authOrder": ["radius", "tacacs", "local"],
            "authFallback": True,
            "adminAuthOrder": True,
            "auditDisable": False,
            "accounting": False,
        }
        self.p_aaa = {
            "authOrder": ["radius", "local"],
            "authFallback": True,
            "adminAuthOrder": True,
            "auditDisable": True,
            "accounting": True,
        }
        self.aaa_dataclass = create_dataclass(TenantAAA, self.aaa)
        self.p_aaa_dataclass = create_dataclass(TenantAAA, self.p_aaa)

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.Session")
    @patch("requests.Response")
    def test_del_aaa(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.get_data.return_value = self.aaa
        mock_session.delete.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = TenantAaaAPI(mock_session).del_aaa()
        # Assert
        self.assertEqual(answer, expected_outcome)

    @parameterized.expand([[200, True], [400, False]])
    @patch("requests.Response")
    @patch("catalystwan.session.Session")
    def test_add_aaa(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.get_data.return_value = self.aaa
        mock_session.post.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = TenantAaaAPI(mock_session).add_aaa(self.aaa_dataclass)
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("catalystwan.session.Session")
    def test_get_aaa(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.aaa
        # Act
        answer = TenantAaaAPI(mock_session).get_aaa()
        # Assert
        self.assertEqual(answer, self.aaa_dataclass)

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.Session")
    @patch("requests.Response")
    def test_put_aaa(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.put.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = TenantAaaAPI(mock_session).put_aaa(self.p_aaa_dataclass)
        # Assert
        self.assertEqual(answer, expected_outcome)


class TestRadiusAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.radius_server_list = [
            {
                "address": "10.0.5.143",
                "authPort": 1812,
                "acctPort": 1813,
                "vpn": 1,
                "vpnIpSubnet": "192.168.1.0/24",
                "key": "testing",
                "secretKey": str(uuid4()),
                "priority": 1,
            },
            {
                "address": "10.0.5.144",
                "authPort": 1812,
                "acctPort": 1813,
                "vpn": 1,
                "vpnIpSubnet": "192.168.1.0/24",
                "key": "testing",
                "secretKey": str(uuid4()),
                "priority": 1,
            },
            {
                "address": "10.0.5.145",
                "authPort": 1812,
                "acctPort": 1813,
                "vpn": 1,
                "vpnIpSubnet": "192.168.1.0/24",
                "key": "testing",
                "secretKey": str(uuid4()),
                "priority": 1,
            },
        ]

        self.p_radius_server_list = [
            {
                "address": "10.0.5.143",
                "authPort": 1812,
                "acctPort": 1813,
                "vpn": 1,
                "vpnIpSubnet": "192.168.1.0/24",
                "key": "testing",
                "secretKey": str(uuid4()),
                "priority": 1,
            },
            {
                "address": "10.0.5.144",
                "authPort": 1812,
                "acctPort": 1813,
                "vpn": 1,
                "vpnIpSubnet": "192.168.1.0/24",
                "key": "testing",
                "secretKey": str(uuid4()),
                "priority": 1,
            },
        ]
        self.radius_server = {"timeout": 5, "retransmit": 3, "server": self.radius_server_list}
        self.p_radius_server = {"timeout": 5, "retransmit": 3, "server": self.p_radius_server_list}
        self.radius_server_dataclass = create_dataclass(TenantRadiusServer, self.radius_server)
        self.p_radius_server_dataclass = create_dataclass(TenantRadiusServer, self.p_radius_server)

    @parameterized.expand([[204, True], [400, False]])
    @patch("catalystwan.session.Session")
    @patch("requests.Response")
    def test_del_radius(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.get_data.return_value = self.radius_server
        mock_session.delete.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = TenantRadiusAPI(mock_session).delete_radius()
        # Assert
        self.assertEqual(answer, expected_outcome)

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.Session")
    @patch("requests.Response")
    def test_add_radius(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.get_data.return_value = self.radius_server
        mock_session.post.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = TenantRadiusAPI(mock_session).add_radius(self.radius_server_dataclass)
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("catalystwan.session.Session")
    def test_get_radius(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.radius_server
        # Act
        answer = TenantRadiusAPI(mock_session).get_radius()
        # Assert
        self.assertEqual(answer, self.radius_server_dataclass)

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.Session")
    @patch("requests.Response")
    def test_put_radius(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.put.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = TenantRadiusAPI(mock_session).put_radius(self.p_radius_server_dataclass)
        # Assert
        self.assertEqual(answer, expected_outcome)


class TestTacacsAPI(unittest.TestCase):
    def setUp(self) -> None:
        self.tacacs_server_list = [
            {
                "address": "10.0.5.141",
                "authPort": 49,
                "vpn": 1,
                "vpnIpSubnet": "192.168.1.0/24",
                "key": "testing",
                "secretKey": str(uuid4()),
                "priority": 1,
            },
            {
                "address": "10.0.5.142",
                "authPort": 49,
                "vpn": 1,
                "vpnIpSubnet": "192.168.1.0/24",
                "key": "testing",
                "secretKey": str(uuid4()),
                "priority": 2,
            },
            {
                "address": "10.0.5.151",
                "authPort": 49,
                "vpn": 1,
                "vpnIpSubnet": "192.168.1.0/24",
                "key": "testing",
                "secretKey": str(uuid4()),
                "priority": 3,
            },
        ]

        self.p_tacacs_server_list = [
            {
                "address": "10.0.5.141",
                "authPort": 49,
                "vpn": 1,
                "vpnIpSubnet": "192.168.1.0/24",
                "key": "testing",
                "secretKey": str(uuid4()),
                "priority": 1,
            },
            {
                "address": "10.0.5.142",
                "authPort": 49,
                "vpn": 1,
                "vpnIpSubnet": "192.168.1.0/24",
                "key": "testing",
                "secretKey": str(uuid4()),
                "priority": 2,
            },
        ]
        self.tacacs_server = {"timeout": 5, "authentication": "PAP", "server": self.tacacs_server_list}
        self.p_tacacs_server = {"timeout": 5, "retransmit": "ASCII", "server": self.p_tacacs_server_list}
        self.tacacs_server_dataclass = create_dataclass(TenantTacacsServer, self.tacacs_server)
        self.p_tacacs_server_dataclass = create_dataclass(TenantTacacsServer, self.p_tacacs_server)

    @parameterized.expand([[204, True], [400, False]])
    @patch("catalystwan.session.Session")
    @patch("requests.Response")
    def test_del_tacacs(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.get_data.return_value = self.tacacs_server
        mock_session.delete.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = TenantTacacsAPI(mock_session).delete_tacacs()
        # Assert
        self.assertEqual(answer, expected_outcome)

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.Session")
    @patch("requests.Response")
    def test_add_tacacs(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.get_data.return_value = self.tacacs_server
        mock_session.post.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = TenantTacacsAPI(mock_session).add_tacacs(self.tacacs_server_dataclass)
        # Assert
        self.assertEqual(answer, expected_outcome)

    @patch("catalystwan.session.Session")
    def test_get_tacacs(self, mock_session):
        # Arrange
        mock_session.get_data.return_value = self.tacacs_server
        # Act
        answer = TenantTacacsAPI(mock_session).get_tacacs()
        # Assert
        self.assertEqual(answer, self.tacacs_server_dataclass)

    @parameterized.expand([[200, True], [400, False]])
    @patch("catalystwan.session.Session")
    @patch("requests.Response")
    def test_put_tacacs(self, status_code, expected_outcome, mock_session, mock_response):
        # Arrange
        mock_session.put.return_value = mock_response
        mock_response.status_code = status_code
        # Act
        answer = TenantTacacsAPI(mock_session).put_tacacs(self.p_tacacs_server_dataclass)
        # Assert
        self.assertEqual(answer, expected_outcome)



# Copyright 2022 Cisco Systems, Inc. and its affiliates

from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from tenacity import retry, retry_if_result, stop_after_attempt, wait_fixed  # type: ignore

from catalystwan.api.basic_api import DevicesAPI, DeviceStateAPI
from catalystwan.dataclasses import Device
from catalystwan.utils.certificate_status import CertificateStatus
from catalystwan.utils.operation_status import OperationStatus
from catalystwan.utils.personality import Personality
from catalystwan.utils.reachability import Reachability
from catalystwan.utils.validate_status import ValidateStatus

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from catalystwan.session import ManagerSession


class DeviceActionAPI(ABC):
    """API method to execute action on Device."""

    def __init__(self, session: ManagerSession, dev: Device):
        self.session = session
        self.dev = dev
        self.action_status = ""
        self.action_id = ""
        self.action_status_api = "/dataservice/device/action/status/"

    def __str__(self):
        return str(self.session)

    @abstractmethod
    def execute(self):
        raise NotImplementedError

    @abstractmethod
    def wait_for_completed(
        self,
        sleep_seconds: int,
        timeout_seconds: int,
        expected_status: str,
        expected_reachability: str,
    ):
        raise NotImplementedError


class RebootAction(DeviceActionAPI):
    """API method to perform reboot on Device."""

    def execute(self):
        """Reboots the device.

        Returns:
            reboot process id

        Raises:
            Exception when reboot was not successful.

        Usage example:
        # Create session and chose device
        session = create_manager_session(...)
        device = DevicesAPI(session).get().filter(personality = Personality.VSMART)[0]
        # Restart device
        RebootAction(session, device).execute()
        """
        controllers = (Personality.VBOND, Personality.VSMART)
        device_type = "controller" if self.dev.personality in controllers else self.dev.personality.value

        body = {
            "action": "reboot",
            "deviceType": device_type,
            "devices": [{"deviceIP": self.dev.id, "deviceId": self.dev.uuid}],
        }
        response = self.session.post("/dataservice/device/action/reboot", json=body).json()
        if response.get("id"):
            self.action_id = response["id"]
        else:
            raise Exception(f"Problem with reboot of {self.dev.id} occurred")

    def wait_for_completed(
        self,
        sleep_seconds: int = 15,
        timeout_seconds: int = 1800,
        expected_status: str = OperationStatus.SUCCESS.value,
        expected_reachability: str = Reachability.REACHABLE.value,
    ):
        def check_status(action_data):
            return not action_data == expected_status

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check_status),
        )
        def wait_for_come_up():
            status_api = f"{self.action_status_api}{self.action_id}"
            # from my observation it is necessary to wait minimum 1 second for data related with reboot
            try:
                action_data = self.session.get_data(f"{status_api}")[0]["status"]
                logger.debug(f"Status of device {self.dev.hostname} reboot is: {action_data}")
            except IndexError:
                action_data = ""
            status = DeviceStateAPI(self.session).get_system_status(self.dev.id)
            # it is necessary to wait also for Success of reboot because device can be reachable even several
            # seconds after execute reboot
            if status.reachability.value == expected_reachability:
                return action_data
            else:
                return None

        wait_for_come_up()


class ValidateAction(DeviceActionAPI):  # TODO check
    """
    API method to perform validate Device

    Usage example:
    # Create session and chose device
    session = create_manager_session(...)
    device = DevicesAPI(session).get().filter(personality = Personality.VSMART)[0]
    # Validate device
    ValidateAction(session, device).execute()
    """

    def execute(self, valid: bool = True):
        """
        validate device
        """
        body = {
            "chasisNumber": self.dev.uuid,
            "serialNumber": self.dev.board_serial,
            "validity": "valid" if valid else "invalid",
        }

        response = self.session.post(url="/dataservice/certificate/save/vedge/list", json=body).json()
        if response.get("id"):
            self.action_id = response["id"]
        else:
            raise Exception(f"Problem with validate of {self.dev.id} occurred")

    def wait_for_completed(
        self,
        sleep_seconds: int = 15,
        timeout_seconds: int = 180,
        expected_status: str = ValidateStatus.validate.value,
        expected_reachability: str = Reachability.REACHABLE.value,
    ):
        def check_status(action_data):
            return not action_data["validity"] == expected_status

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check_status),
        )
        def wait_for_come_up():
            status = DeviceStateAPI(self.session).get_system_status(self.dev.id)
            if status.reachability.value == expected_reachability:
                return self.session.get_data(f"/dataservice/device?host-name={self.dev.hostname}")[0]
            else:
                return None

        wait_for_come_up()


class DecommissionAction(DeviceActionAPI):
    """
    API method to decommission Device

    Usage example:
    # Create session and chose device
    session = create_manager_session(...)
    device = DevicesAPI(session).get().filter(personality = Personality.VSMART)[0]
    # Decommission device
    DecommissionAction(session, device).execute()
    """

    def execute(self):
        """
        Decommission device.
        """
        url = f"/dataservice/system/device/decommission/{self.dev.uuid}"
        response = self.session.put(url)
        if response.status_code != 200:
            raise Exception(f"Problem with decomission of {self.dev.id} occurred")

    def wait_for_completed(
        self,
        sleep_seconds: int = 15,
        timeout_seconds: int = 20,
        expected_status: str = CertificateStatus.generated.value,
        expected_reachability=Reachability.UNREACHABLE.value,
    ):
        def check_status(action_data):
            return not action_data.vedgeCertificateState == expected_status

        @retry(
            wait=wait_fixed(sleep_seconds),
            stop=stop_after_attempt(int(timeout_seconds / sleep_seconds)),
            retry=retry_if_result(check_status),
        )
        def wait_for_status():
            status = DevicesAPI(self.session).get_device_details(self.dev.uuid)
            if status.reachability.value == expected_reachability:
                return status
            else:
                return None

        wait_for_status()



# Copyright 2022 Cisco Systems, Inc. and its affiliates

import logging
from unittest import TestCase
from unittest.mock import patch

from catalystwan.api.alarms_api import AlarmsAPI, AlarmVerification
from catalystwan.dataclasses import AlarmData
from catalystwan.typed_list import DataSequence
from catalystwan.utils.alarm_status import Severity
from catalystwan.utils.creation_tools import create_dataclass, flatten_dict


class TestAlarmsAPI(TestCase):
    def setUp(self) -> None:
        self.alarms_data = {
            "data": [
                {
                    "type": "site_up",
                    "component": "OMP",
                    "severity": "Critical",
                    "active": False,
                    "acknowledged": False,
                    "consumed_events": [
                        {
                            "eventname": "omp-peer-state-change",
                            "peer-new-state": "handshake-in-gr",
                            "eventId": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
                            "builtBy": "EventDataCollector",
                            "component": "OMP",
                            "severity-level": "major ",
                            "vmanage-system-ip": "1.1.1.1",
                            "entry_time": 1671028426000,
                            "system-ip": "1.1.1.1",
                            "host-name": "vm1",
                            "peer": "1.1.1.1",
                        }
                    ],
                    "site_id": "1",
                    "devices": [{"system-ip": "1.1.1.1"}],
                },
                {
                    "type": "site_up",
                    "component": "OMP",
                    "severity": "Major",
                    "active": False,
                    "acknowledged": False,
                    "consumed_events": [
                        {
                            "eventname": "omp-peer-state-change",
                            "peer-new-state": "handshake-in-gr",
                            "eventId": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaab",
                            "builtBy": "EventDataCollector",
                            "component": "OMP",
                            "severity-level": "major ",
                            "vmanage-system-ip": "1.1.1.2",
                            "entry_time": 1671028426000,
                            "system-ip": "1.1.1.2",
                            "host-name": "vm2",
                            "peer": "1.1.1.2",
                        }
                    ],
                    "site_id": "2",
                    "devices": [{"system-ip": "1.1.1.2"}],
                },
                {
                    "type": "site_up",
                    "component": "OMP",
                    "severity": "Medium",
                    "active": False,
                    "acknowledged": False,
                    "consumed_events": [
                        {
                            "eventname": "omp-peer-state-change",
                            "peer-new-state": "handshake-in-gr",
                            "eventId": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaac",
                            "builtBy": "EventDataCollector",
                            "component": "OMP",
                            "severity-level": "major ",
                            "vmanage-system-ip": "1.1.1.3",
                            "entry_time": 1671028426000,
                            "system-ip": "1.1.1.3",
                            "host-name": "vm3",
                            "peer": "1.1.1.3",
                        }
                    ],
                    "site_id": "3",
                    "devices": [{"system-ip": "1.1.1.3"}],
                },
                {
                    "type": "site_up",
                    "component": "OMP",
                    "severity": "Minor",
                    "active": False,
                    "acknowledged": False,
                    "consumed_events": [
                        {
                            "eventname": "omp-peer-state-change",
                            "peer-new-state": "handshake-in-gr",
                            "eventId": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaad",
                            "builtBy": "EventDataCollector",
                            "component": "OMP",
                            "severity-level": "major ",
                            "vmanage-system-ip": "1.1.1.4",
                            "entry_time": 1671028426000,
                            "system-ip": "1.1.1.4",
                            "host-name": "vm4",
                            "peer": "1.1.1.4",
                        }
                    ],
                    "site_id": "4",
                    "devices": [{"system-ip": "1.1.1.4"}],
                },
            ]
        }
        self.alarms = self.alarms_data["data"]
        self.critical_alarms_data = {"data": [self.alarms[0]]}
        self.major_alarms_data = {"data": [self.alarms[1]]}
        self.medium_alarms_data = {"data": [self.alarms[2]]}
        self.minor_alarms_data = {"data": [self.alarms[3]]}
        self.alarms_dataseq = DataSequence(
            AlarmData, [create_dataclass(AlarmData, flatten_dict(alarm)) for alarm in self.alarms]
        )
        self.critical_alarms_dataseq = DataSequence(
            AlarmData, [create_dataclass(AlarmData, flatten_dict(self.alarms[0]))]
        )
        self.major_alarms_dataseq = DataSequence(AlarmData, [create_dataclass(AlarmData, flatten_dict(self.alarms[1]))])
        self.medium_alarms_dataseq = DataSequence(
            AlarmData, [create_dataclass(AlarmData, flatten_dict(self.alarms[2]))]
        )
        self.minor_alarms_dataseq = DataSequence(AlarmData, [create_dataclass(AlarmData, flatten_dict(self.alarms[3]))])
        self.maxDiff = None

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_alarms(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.json.return_value = self.alarms_data
        # Act
        answer = AlarmsAPI(mock_session).get()
        # Assert
        self.assertEqual(answer, self.alarms_dataseq)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_critical_alarms(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.json.return_value = self.alarms_data
        # Act
        answer = AlarmsAPI(mock_session).get().filter(severity=Severity.CRITICAL)
        # Assert
        self.assertEqual(answer, self.critical_alarms_dataseq)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_major_alarms(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.json.return_value = self.alarms_data
        # Act
        answer = AlarmsAPI(mock_session).get().filter(severity=Severity.MAJOR)
        # Assert
        self.assertEqual(answer, self.major_alarms_dataseq)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_medium_alarms(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.json.return_value = self.alarms_data
        # Act
        answer = AlarmsAPI(mock_session).get().filter(severity=Severity.MEDIUM)
        # Assert
        self.assertEqual(answer, self.medium_alarms_dataseq)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_minor_alarms(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.json.return_value = self.alarms_data
        # Act
        answer = AlarmsAPI(mock_session).get().filter(severity=Severity.MINOR)
        # Assert
        self.assertEqual(answer, self.minor_alarms_dataseq)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_get_not_viewed_alarms(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.json.return_value = self.alarms_data
        # Act
        answer = AlarmsAPI(mock_session).get().filter(viewed=False)
        # Assert
        self.assertEqual(answer, self.alarms_dataseq)

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_check_alarms(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.json.return_value = self.alarms_data
        # Act
        answer = AlarmsAPI(mock_session).check_alarms(self.minor_alarms_data["data"])
        # Assert
        self.assertEqual(answer, {"found": set(self.minor_alarms_dataseq), "not-found": set()})

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_check_alarms_not_found(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.dataseq.return_value = self.major_alarms_dataseq
        # Act
        answer = AlarmsAPI(mock_session).check_alarms(self.minor_alarms_data["data"], 2, 1)
        # Assert
        self.assertEqual(answer, {"found": set(), "not-found": set(self.minor_alarms_dataseq)})

    # # # test AlarmVerification class
    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_verify(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.dataseq.return_value = self.major_alarms_dataseq
        alarms_getter = AlarmsAPI(mock_session).get
        logger = logging.getLogger("test")
        # Act
        answer = AlarmVerification(logger, alarms_getter)
        answer.verify(set(self.minor_alarms_dataseq), 2, 1)
        # Assert
        self.assertEqual(answer.found, set())
        self.assertEqual(answer.not_found, set(self.minor_alarms_dataseq))

    @patch("catalystwan.response.ManagerResponse")
    @patch("catalystwan.session.ManagerSession")
    def test_verify_not_found(self, mock_session, mock_response):
        # Arrange
        mock_session.post.return_value = mock_response
        mock_response.dataseq.return_value = self.critical_alarms_dataseq
        alarms_getter = AlarmsAPI(mock_session).get
        logger = logging.getLogger("test")
        # Act
        answer = AlarmVerification(logger, alarms_getter)
        answer.verify(set(self.minor_alarms_dataseq), 2, 1)
        # Assert
        self.assertEqual(answer.found, set())
        self.assertEqual(answer.not_found, set(self.minor_alarms_dataseq))



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from typing import Dict, List, Literal, Set, Tuple


def check_fields_exclusive(values: Dict, field_names: Set[str], at_least_one: bool = False) -> bool:
    """Helper method to check fields are mutually exclusive

    Args:
        values (Dict): BaseModel field values
        field_names (Set[str]): set of field names that we want to be mutually exclusive
        at_least_one (bool, optional): Additionaly check if at least one of fields is not None

    Raises:
        ValueError: When fields are not mutually exclusive

    Returns:
        bool: True if at least one field was present
    """
    assigned = [values.get(field_name) for field_name in field_names if values.get(field_name) is not None]
    if len(assigned) == 0 and at_least_one:
        raise ValueError(f"At least one of given fields {field_names} must be assigned")
    if len(assigned) > 1:
        raise ValueError(f"Fields {field_names} are mutually exclusive")
    return True if len(assigned) > 0 else False


def check_any_of_exclusive_field_sets(values: Dict, field_sets: List[Tuple[Set[str], bool]]):
    """This is very specific validator but common in policy definitions model.
    It checks that fields in each of the sets are mutually exclusive and also
    guarantees that at least one of the values is present from all sets.

    Args:
        values (Dict): BaseModel field values
        field_sets (Set[Tuple[Set[str]], bool]): Set of tuples each tuple should
        contain field names set and flag to check if at least one of fields is present within a set

    Raises:
        ValueError: When fields are not mutually exclusive or none of the field values is present

    """
    any_assigned = False
    for field_names, at_least_one in field_sets:
        if check_fields_exclusive(values, field_names, at_least_one):
            any_assigned = True
    if not any_assigned:
        all_sets_field_names = [s[0] for s in field_sets]
        raise ValueError(f"One of {all_sets_field_names} must be assigned")


InterfaceType = Literal[
    "Ethernet",
    "FastEthernet",
    "FiveGigabitEthernet",
    "FortyGigabitEthernet",
    "GigabitEthernet",
    "HundredGigE",
    "Loopback",
    "TenGigabitEthernet",
    "Tunnel",
    "TwentyFiveGigabitEthernet",
    "TwentyFiveGigE",
    "TwoGigabitEthernet",
    "VirtualPortGroup",
    "Vlan",
]

TLOCColor = Literal[
    "default",
    "mpls",
    "metro-ethernet",
    "biz-internet",
    "public-internet",
    "lte",
    "3g",
    "red",
    "green",
    "blue",
    "gold",
    "silver",
    "bronze",
    "custom1",
    "custom2",
    "custom3",
    "private1",
    "private2",
    "private3",
    "private4",
    "private5",
    "private6",
]


WellKnownBGPCommunities = Literal[
    "internet",
    "local-AS",
    "no-advertise",
    "no-export",
]

ServiceChainNumber = Literal[
    "SC1",
    "SC2",
    "SC3",
    "SC4",
    "SC5",
    "SC6",
    "SC7",
    "SC8",
    "SC9",
    "SC10",
    "SC11",
    "SC12",
    "SC13",
    "SC14",
    "SC15",
    "SC16",
]



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Address, IPv6Network
from typing import Any, List, Literal, Set, Tuple, Union
from uuid import UUID

from pydantic import ConfigDict, Field
from typing_extensions import Annotated

from catalystwan.models.policy.policy_definition import (
    ClassMapAction,
    ClassMapListEntry,
    CountAction,
    DefaultAction,
    DefinitionWithSequencesCommonBase,
    DestinationDataIPv6PrefixListEntry,
    DestinationIPv6Entry,
    DestinationPortEntry,
    LogAction,
    Match,
    MirrorAction,
    NextHeaderEntry,
    NextHopEntry,
    PacketLengthEntry,
    PLPEntry,
    PolicerAction,
    PolicyActionType,
    PolicyDefinitionBase,
    PolicyDefinitionSequenceBase,
    Reference,
    SourceDataIPv6PrefixListEntry,
    SourceIPv6Entry,
    SourcePortEntry,
    TCPEntry,
    TrafficClassEntry,
    accept_action,
)

AclIPv6PolicySequenceMatchEntry = Annotated[
    Union[
        ClassMapListEntry,
        DestinationDataIPv6PrefixListEntry,
        DestinationIPv6Entry,
        DestinationPortEntry,
        NextHeaderEntry,
        PacketLengthEntry,
        PLPEntry,
        SourceDataIPv6PrefixListEntry,
        SourceIPv6Entry,
        SourcePortEntry,
        TCPEntry,
        TrafficClassEntry,
    ],
    Field(discriminator="field"),
]

AclIPv6PolicySequenceActions = Any  # TODO


class AclIPv6PolicyHeader(PolicyDefinitionBase):
    type: Literal["aclv6"] = "aclv6"


class AclIPv6PolicySequenceMatch(Match):
    entries: List[AclIPv6PolicySequenceMatchEntry] = []


class AclIPv6PolicySequence(PolicyDefinitionSequenceBase):
    sequence_type: Literal["aclv6"] = Field(
        default="aclv6", serialization_alias="sequenceType", validation_alias="sequenceType"
    )
    base_action: PolicyActionType = Field(
        default="accept", serialization_alias="baseAction", validation_alias="baseAction"
    )
    match: AclIPv6PolicySequenceMatch = AclIPv6PolicySequenceMatch()
    actions: List[AclIPv6PolicySequenceActions] = []
    model_config = ConfigDict(populate_by_name=True)

    def match_next_header(self, next_header: int) -> None:
        self._insert_match(NextHeaderEntry(value=str(next_header)))

    def match_packet_length(self, packet_lengths: Tuple[int, int]) -> None:
        self._insert_match(PacketLengthEntry.from_range(packet_lengths))

    def match_low_plp(self) -> None:
        self._insert_match(PLPEntry(value="low"))

    def match_high_plp(self) -> None:
        self._insert_match(PLPEntry(value="high"))

    def match_source_data_prefix_list(self, data_prefix_list_id: UUID) -> None:
        self._insert_match(SourceDataIPv6PrefixListEntry(ref=data_prefix_list_id))

    def match_source_ip(self, networks: List[IPv6Network]) -> None:
        self._insert_match(SourceIPv6Entry.from_ipv6_networks(networks))

    def match_source_port(self, ports: Set[int] = set(), port_ranges: List[Tuple[int, int]] = []) -> None:
        self._insert_match(SourcePortEntry.from_port_set_and_ranges(ports, port_ranges))

    def match_destination_data_prefix_list(self, data_prefix_list_id: UUID) -> None:
        self._insert_match(DestinationDataIPv6PrefixListEntry(ref=data_prefix_list_id))

    def match_destination_ip(self, networks: List[IPv6Network]) -> None:
        self._insert_match(DestinationIPv6Entry.from_ipv6_networks(networks))

    def match_destination_port(self, ports: Set[int] = set(), port_ranges: List[Tuple[int, int]] = []) -> None:
        self._insert_match(DestinationPortEntry.from_port_set_and_ranges(ports, port_ranges))

    def match_tcp(self) -> None:
        self._insert_match(TCPEntry())

    def match_class_map_list_entry(self, class_map_list_id: UUID) -> None:
        self._insert_match(ClassMapListEntry(ref=class_map_list_id))

    def match_traffic_class(self, traffic_class: int) -> None:
        self._insert_match(TrafficClassEntry(value=str(traffic_class)))

    def associate_count_action(self, counter_name: str) -> None:
        self._insert_action(CountAction(parameter=counter_name))

    def associate_log_action(self) -> None:
        self._insert_action(LogAction())

    @accept_action
    def associate_next_hop_action(self, next_hop: IPv4Address) -> None:
        self._insert_action_in_set(NextHopEntry(value=next_hop))

    @accept_action
    def associate_traffic_class_action(self, traffic_class: int) -> None:
        self._insert_action_in_set(TrafficClassEntry(value=str(traffic_class)))

    @accept_action
    def associate_mirror_action(self, mirror_list_id: UUID) -> None:
        self._insert_action(MirrorAction(parameter=Reference(ref=mirror_list_id)))

    @accept_action
    def associate_class_map_list_action(self, class_map_list_id: UUID) -> None:
        self._insert_action(ClassMapAction(parameter=Reference(ref=class_map_list_id)))

    @accept_action
    def associate_policer_list_action(self, policer_list_id: UUID) -> None:
        self._insert_action(PolicerAction(parameter=Reference(ref=policer_list_id)))


class AclIPv6Policy(AclIPv6PolicyHeader, DefinitionWithSequencesCommonBase):
    sequences: List[AclIPv6PolicySequence] = []
    default_action: DefaultAction = Field(
        default=DefaultAction(type="drop"),
        serialization_alias="defaultAction",
        validation_alias="defaultAction",
    )
    model_config = ConfigDict(populate_by_name=True)

    def add_acl_sequence(
        self, name: str = "Access Control List", base_action: PolicyActionType = "accept"
    ) -> AclIPv6PolicySequence:
        seq = AclIPv6PolicySequence(
            sequence_name=name,
            base_action=base_action,
            sequence_ip_type="ipv6",
        )
        self.add(seq)
        return seq



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from catalystwan.api.templates.models.cisco_banner_model import CiscoBannerModel

banner_model = CiscoBannerModel(  # type: ignore
    template_name="banner_1", template_description="na", login_banner="login banner", motd_banner="motd_bnanner"
)



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv4Network, IPv6Network
from typing import Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, model_validator

from catalystwan.models.common import check_any_of_exclusive_field_sets, check_fields_exclusive
from catalystwan.models.policy.policy_definition import PolicyDefinitionBase, Reference, VariableName

SequenceIPType = Literal[
    "ipv4",
    "ipv6",
]


class SecurityGroupIPv4Definition(BaseModel):
    data_prefix: Union[IPv4Network, VariableName, None] = Field(None, alias="dataPrefix")
    data_prefix_list: Optional[Reference] = Field(None, alias="dataPrefixList")
    fqdn: Optional[str] = None
    fqdn_list: Optional[Reference] = Field(None, alias="fqdnList")
    geo_location: Optional[str] = Field(None, alias="geoLocation")
    geo_location_list: Optional[Reference] = Field(None, alias="geoLocationList")
    port: Optional[str] = None
    port_list: Optional[Reference] = Field(None, alias="portList")
    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    @model_validator(mode="after")
    def check_exclusive_fields(self):
        check_any_of_exclusive_field_sets(
            self.__dict__,
            [
                ({"data_prefix", "data_prefix_list"}, False),
                ({"fqdn", "fqdn_list"}, False),
                ({"geo_location", "geo_location_list"}, False),
                ({"port", "port_list"}, False),
            ],
        )
        return self


class SecurityGroupIPv6Definition(BaseModel):
    data_ipv6_prefix: Union[IPv6Network, VariableName, None] = Field(None, alias="dataIPV6Prefix")
    data_ipv6_prefix_list: Optional[Reference] = Field(None, alias="dataIPV6PrefixList")
    model_config = ConfigDict(extra="forbid", populate_by_name=True)

    @model_validator(mode="after")
    def check_exclusive_fields(self):
        check_fields_exclusive(self.__dict__, {"data_ipv6_prefix", "data_ipv6_prefix_list"}, True)
        return self


class SecurityGroup(PolicyDefinitionBase):
    # TODO: cannot use sequence_ip_type discriminated unions here as this is root of model
    # EndpointAPI would need to support annotated union as payload parameter
    type: Literal["securityGroup"] = "securityGroup"
    mode: Literal["unified"] = "unified"
    sequence_ip_type: SequenceIPType = Field(alias="sequenceIpType")
    definition: Union[SecurityGroupIPv4Definition, SecurityGroupIPv6Definition]

    @model_validator(mode="after")
    def validate_by_sequence_ip_type(self):
        if (
            self.sequence_ip_type == SequenceIPType.IPV4 and isinstance(self.definition, SecurityGroupIPv6Definition)
        ) or (
            self.sequence_ip_type == SequenceIPType.IPV6 and isinstance(self.definition, SecurityGroupIPv4Definition)
        ):
            raise ValueError(f"Incompatible definition for {self.sequence_ip_type} sequence")
        return self



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"
from datetime import datetime
from typing import Any, List, Optional

from packaging.version import Version  # type: ignore
from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, get


class VersionField(Version):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        return Version(value)


class ServerInfo(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    server: Optional[str] = None
    tenancy_mode: Optional[str] = Field(default=None, serialization_alias="tenancyMode", validation_alias="tenancyMode")
    user_mode: Optional[str] = Field(default=None, serialization_alias="userMode", validation_alias="userMode")
    vsession_id: Optional[str] = Field(default=None, serialization_alias="VSessionId", validation_alias="VSessionId")
    is_saml_user: Optional[bool] = Field(default=None, serialization_alias="isSamlUser", validation_alias="isSamlUser")
    is_rbac_vpn_user: Optional[bool] = Field(
        default=None, serialization_alias="isRbacVpnUser", validation_alias="isRbacVpnUser"
    )
    vpns: List[Any] = []
    csrf_token: Optional[str] = Field(default=None, serialization_alias="CSRFToken", validation_alias="CSRFToken")
    provider_domain: Optional[str] = Field(
        default=None, serialization_alias="providerDomain", validation_alias="providerDomain"
    )
    tenant_id: Optional[str] = Field(default=None, serialization_alias="tenantId", validation_alias="tenantId")
    provider_id: Optional[str] = Field(default=None, serialization_alias="providerId", validation_alias="providerId")
    view_mode: Optional[str] = Field(default=None, serialization_alias="viewMode", validation_alias="viewMode")
    capabilities: List[str] = []
    user: Optional[str] = None
    description: Optional[str] = None
    locale: Optional[str] = None
    roles: List[str] = []
    external_user: Optional[bool] = Field(
        default=None, serialization_alias="externalUser", validation_alias="externalUser"
    )
    platform_version: str = Field(default="", serialization_alias="platformVersion", validation_alias="platformVersion")
    general_template: Optional[bool] = Field(
        default=None, serialization_alias="generalTemplate", validation_alias="generalTemplate"
    )
    disable_full_config_push: Optional[bool] = Field(
        default=None, serialization_alias="disableFullConfigPush", validation_alias="disableFullConfigPush"
    )
    enable_server_events: Optional[bool] = Field(
        default=None, serialization_alias="enableServerEvents", validation_alias="enableServerEvents"
    )
    cloudx: Optional[str] = None
    reverseproxy: Optional[str] = None
    vmanage_mode: Optional[str] = Field(default=None, serialization_alias="vmanageMode", validation_alias="vmanageMode")


class AboutInfo(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    title: Optional[str]
    version: str = Field(default="")
    application_version: Optional[str] = Field(
        serialization_alias="applicationVersion", validation_alias="applicationVersion"
    )
    application_server: Optional[str] = Field(
        serialization_alias="applicationServer", validation_alias="applicationServer"
    )
    copyright: Optional[str]
    time: Optional[datetime]
    time_zone: Optional[str] = Field(serialization_alias="timeZone", validation_alias="timeZone")
    logo: Optional[str]


class ServerReady(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    is_server_ready: bool = Field(serialization_alias="isServerReady", validation_alias="isServerReady")


class Client(APIEndpoints):
    @get("/client/server", "data")
    def server(self) -> ServerInfo:
        ...

    @get("/client/server/ready")
    def server_ready(self) -> ServerReady:
        ...

    @get("/client/about", "data")
    def about(self) -> AboutInfo:
        ...



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import Any, Dict, List, Optional, Union

from catalystwan.api.templates.device_variable import DeviceVariable


def find_template_values(
    template_definition: dict,
    templated_values: Optional[dict] = None,
    target_key: str = "vipType",
    target_key_value_to_ignore: str = "ignore",
    target_key_for_template_value: str = "vipValue",
    path: Optional[List[str]] = None,
) -> Dict[str, Union[str, list, dict]]:
    """Based on provided template definition generates a dictionary with template fields and values

    Args:
        template_definition: template definition provided as dict
        templated_values: dictionary, empty at the beginning and filed out with names of fields as keys
            and values of those fields as values
        target_key: name of the key specifying if field is used in template, defaults to 'vipType'
        target_key_value_to_ignore: value of the target key indicating
            that field is not used in template, defaults to 'ignore'
        target_key_for_template_value: name of the key specifying value of field used in template,
            defaults to 'vipValue'
        path: a list of keys indicating current path, defaults to None
    Returns:
        templated_values: dictionary containing template fields as key and values assigned to those fields as values
    """
    if path is None:
        path = []
    if templated_values is None:
        templated_values = {}
    # if value object is reached, try to extract the value
    if target_key in template_definition:
        if template_definition[target_key] == target_key_value_to_ignore:
            return templated_values

        value = template_definition[target_key]  # vipType
        template_value = template_definition.get(target_key_for_template_value)

        field_key = path[-1]
        if value == "variableName":
            # For example this is the current dictionary:
            # field_key is "dns-addr"
            # {
            #     "vipType": "variableName",
            #     "vipValue": "",
            #     "vipObjectType": "object",
            #     "vipVariableName": "vpn_dns_primary",
            # }
            # vipType is "variableName" so we need to return
            # {"dns-addr": DeviceVariable(name="vpn_dns_primary")}
            templated_values[field_key] = DeviceVariable(name=template_definition["vipVariableName"])
            return templated_values

        if template_value is None:
            return template_definition

        if template_definition["vipObjectType"] == "list":
            current_nesting = get_nested_dict(templated_values, path[:-1])
            current_nesting[field_key] = []
            for item in template_value:
                current_nesting[field_key].append(process_list_value(item))
        elif template_definition["vipObjectType"] != "tree":
            current_nesting = get_nested_dict(templated_values, path[:-1])
            current_nesting[field_key] = template_value
        elif isinstance(template_value, dict):
            find_template_values(value, templated_values, path=path)
        elif isinstance(template_value, list):
            current_nesting = get_nested_dict(templated_values, path[:-1])
            current_nesting[field_key] = []
            for item in template_value:
                item_value = find_template_values(item, {})
                if item_value:
                    current_nesting[field_key].append(item_value)
        return templated_values

    # iterate the dict to extract values and assign them to their fields
    for key, value in template_definition.items():
        if isinstance(value, dict) and value != target_key_value_to_ignore:
            find_template_values(value, templated_values, path=path + [key])
    return templated_values


def get_nested_dict(d: dict, path: List[str], populate: bool = True):
    current_dict = d
    for path_key in path:
        if path_key not in current_dict and populate:
            current_dict[path_key] = {}
        current_dict = current_dict[path_key]
    return current_dict


def process_list_value(item: Any, target_key: str = "vipType", target_key_for_template_value: str = "vipValue"):
    if isinstance(item, dict):
        if target_key in item:
            if item["vipObjectType"] == "list":
                result = []
                for nested_item in item[target_key_for_template_value]:
                    result.append(process_list_value(nested_item))
                return result
            elif item["vipObjectType"] == "tree":
                return find_template_values(item[target_key_for_template_value])
            else:
                return item[target_key_for_template_value]
        else:
            return find_template_values(item)
    else:
        return item



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="annotation-unchecked"
import json
import tempfile
import unittest
from enum import Enum
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union
from unittest.mock import MagicMock
from uuid import UUID, uuid4

import pytest  # type: ignore
from packaging.version import Version  # type: ignore
from parameterized import parameterized  # type: ignore
from pydantic import BaseModel, Field
from typing_extensions import Annotated

from catalystwan.endpoints import (
    BASE_PATH,
    JSON,
    APIEndpoints,
    CustomPayloadType,
    PreparedPayload,
    TypeSpecifier,
    delete,
    get,
)
from catalystwan.endpoints import logger as endpoints_logger
from catalystwan.endpoints import post, put, request, versions, view
from catalystwan.exceptions import APIEndpointError, APIRequestPayloadTypeError, APIVersionError, APIViewError
from catalystwan.typed_list import DataSequence
from catalystwan.utils.session_type import ProviderAsTenantView, ProviderView, TenantView


class BaseModelExample(BaseModel):
    id: str
    size: int
    capacity: float
    active: bool


class BaseModelExample2(BaseModel):
    other_id: str
    active: bool


class ParamsModelExample(BaseModel):
    name: str
    color: str


class CustomTypeExample(CustomPayloadType):
    def __init__(self, filename: Path):
        self.filename = filename

    def prepared(self) -> PreparedPayload:
        data = open(self.filename, "rb")
        return PreparedPayload(files={"file": (Path(data.name).name, data)})


class TestAPIEndpoints(unittest.TestCase):
    def setUp(self):
        self.base_path = BASE_PATH
        self.session_mock = MagicMock()
        self.session_mock.request = MagicMock(return_value=MagicMock())
        self.session_mock.api_version = None
        self.session_mock.session_type = None
        self.endpoints = APIEndpoints(self.session_mock)
        self.dict_payload = {
            "id": "XYZ-189",
            "size": 100,
            "capacity": 1.7,
            "active": True,
        }
        self.json_payload = json.dumps(self.dict_payload)
        self.basemodel_payload = BaseModelExample.model_validate(self.dict_payload)
        self.list_dict_payload = [self.dict_payload] * 2
        self.dict_params = {
            "name": "purple",
            "color": "haze",
        }
        self.basemodel_params = ParamsModelExample.model_validate(self.dict_params)
        self.basemodel_sequence_payload = [self.basemodel_payload] * 2

    @parameterized.expand(
        [
            ("<2.0", "1.9.9"),
            (">=1.9, !=3.0", "3.1"),
            (">0.9, <=1.3", "1.3"),
        ]
    )
    def test_versions_decorator_passes(self, supported_versions, current_version):
        class ExampleAPI(APIEndpoints):
            @versions(supported_versions=supported_versions, raises=True)
            def versions_decorated_method(self):
                pass

        self.session_mock.api_version = Version(current_version)
        api = ExampleAPI(self.session_mock)
        api.versions_decorated_method()

    @parameterized.expand(
        [
            ("<2.0", "2.0"),
            ("<=1.9, !=3.0", "3.0"),
            (">0.9, <1.3", "1.3"),
        ]
    )
    def test_versions_decorator_raises(self, supported_versions, current_version):
        class ExampleAPI(APIEndpoints):
            @versions(supported_versions=supported_versions, raises=True)
            def versions_decorated_method(self):
                pass

        self.session_mock.api_version = Version(current_version)
        api = ExampleAPI(self.session_mock)
        with self.assertRaises(APIVersionError):
            api.versions_decorated_method()

    def test_version_decorator_logs_warning(self):
        supported_versions = "<1.6"
        current_version = "1.7"

        class ExampleAPI(APIEndpoints):
            @versions(supported_versions=supported_versions, raises=False)
            def versions_decorated_method(self):
                pass

        self.session_mock.api_version = Version(current_version)
        api = ExampleAPI(self.session_mock)
        with self.assertLogs(endpoints_logger, level="WARNING") as log:
            api.versions_decorated_method()
            assert supported_versions in log.output[0]
            assert current_version in log.output[0]

    @parameterized.expand(
        [
            ({ProviderView}, ProviderView),
            ({TenantView, ProviderAsTenantView}, TenantView),
            ({ProviderAsTenantView}, ProviderAsTenantView),
        ]
    )
    def test_view_decorator_passes(self, allowed_sessions, current_session):
        class ExampleAPI(APIEndpoints):
            @view(allowed_session_types=allowed_sessions, raises=True)
            def versions_decorated_method(self):
                pass

        self.session_mock.session_type = current_session
        api = ExampleAPI(self.session_mock)
        api.versions_decorated_method()

    @parameterized.expand(
        [
            ({ProviderView}, ProviderAsTenantView),
            ({TenantView, ProviderAsTenantView}, ProviderView),
            ({ProviderAsTenantView}, TenantView),
        ]
    )
    def test_view_decorator_raises(self, allowed_sessions, current_session):
        class ExampleAPI(APIEndpoints):
            @view(allowed_session_types=allowed_sessions, raises=True)
            def versions_decorated_method(self):
                pass

        self.session_mock.session_type = current_session
        api = ExampleAPI(self.session_mock)
        with self.assertRaises(APIViewError):
            api.versions_decorated_method()

    def test_view_decorator_logs_warinig(self):
        allowed_sessions = {ProviderAsTenantView, TenantView}
        current_session = ProviderView

        class ExampleAPI(APIEndpoints):
            @view(allowed_session_types=allowed_sessions)
            def versions_decorated_method(self):
                pass

        self.session_mock.session_type = current_session
        api = ExampleAPI(self.session_mock)
        with self.assertLogs(endpoints_logger, level="WARNING") as log:
            api.versions_decorated_method()
            for allowed in allowed_sessions:
                assert str(allowed) in log.output[0]
            assert str(current_session) in log.output[0]

    @pytest.mark.filterwarnings("ignore::DeprecationWarning")
    def test_basemodel_payload(self):
        self.endpoints._request("GET", f"/{__name__}", payload=self.basemodel_payload)
        _, kwargs = self.session_mock.request.call_args
        assert json.loads(kwargs.get("data")) == self.dict_payload

    @pytest.mark.filterwarnings("ignore::DeprecationWarning")
    def test_basemodel_sequence_payload(self):
        self.endpoints._request("GET", f"/{__name__}", payload=self.basemodel_sequence_payload)
        _, kwargs = self.session_mock.request.call_args
        assert json.loads(kwargs.get("data")) == self.list_dict_payload

    def test_custom_payload(self):
        expected_data = b"\xFFsomething\x00"
        with tempfile.TemporaryDirectory() as tempdir:
            tempdir_path = Path(tempdir)
            payload_file = tempdir_path / "payload.bin"
            with open(payload_file, "wb") as tmpfile:
                tmpfile.write(expected_data)
            custom_payload = CustomTypeExample(payload_file)
            self.endpoints._request("POST", f"/{__name__}", payload=custom_payload)
            _, kwargs = self.session_mock.request.call_args
            prepared = custom_payload.prepared()
            assert kwargs.get("data") == prepared.data
            assert kwargs.get("headers") == prepared.headers
            assert kwargs.get("files")["file"][0] == prepared.files["file"][0]

    def test_dict_payload(self):
        self.endpoints._request("POST", f"/{__name__}", payload=self.dict_payload)
        _, kwargs = self.session_mock.request.call_args
        assert kwargs.get("data") == self.json_payload

    @parameterized.expand(
        [
            ("JStr", str),
            (1, int),
            (3.33, float),
            (True, bool),
            ({"a": {"b": [1.6]}}, dict),
            ([1, 2, {"a": {"b": [1.6]}}], list),
        ]
    )
    def test_forced_json_payload(self, payload, _):
        self.endpoints._request("POST", f"/{__name__}", payload=payload, force_json_payload=True)
        _, kwargs = self.session_mock.request.call_args
        assert kwargs.get("data") == json.dumps(payload)

    def test_str_payload(self):
        str_payload = "This is string payload!"
        self.endpoints._request("POST", f"/{__name__}", payload=str_payload)
        _, kwargs = self.session_mock.request.call_args
        assert kwargs.get("data") == str_payload

    def test_bytes_payload(self):
        bytes_payload = b"\xEFtest\x00"
        self.endpoints._request("POST", f"/{__name__}", payload=bytes_payload)
        _, kwargs = self.session_mock.request.call_args
        assert kwargs.get("data") == bytes_payload

    def test_unexpected_payload(self):
        with self.assertRaises(APIRequestPayloadTypeError):
            self.endpoints._request("GET", f"/{__name__}", payload={1, 2, 3})

    def test_dict_params(self):
        self.endpoints._request("POST", f"/{__name__}", params=self.dict_params)
        _, kwargs = self.session_mock.request.call_args
        assert kwargs.get("params") == self.dict_params

    @pytest.mark.filterwarnings("ignore::DeprecationWarning")
    def test_basemodel_params(self):
        self.endpoints._request("POST", f"/{__name__}", params=self.basemodel_params)
        _, kwargs = self.session_mock.request.call_args
        assert kwargs.get("params") == self.dict_params

    def test_request_decorator_forbidden_url_field_name(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("GET", "/v1/data/{payload}")
                def get_data(self, payload: str) -> None:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_unsupported_return_type(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("GET", "/v1/data/")
                def get_data(self) -> DataSequence:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_unsupported_datasequence_return_type(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("GET", "/v1/data/")
                def get_data(self) -> DataSequence[str]:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_unsupported_composite_return_type(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("GET", "/v1/data")
                def get_data(self) -> List[BaseModelExample]:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_unsupported_payload_type(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("POST", "/v1/data")
                def get_data(self, payload: DataSequence) -> None:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_unsupported_payload_sequence_type(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("POST", "/v1/data")
                def get_data(self, payload: List[str]) -> None:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_unsupported_payload_composite_type(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("POST", "/v1/data")
                def get_data(self, payload: Dict[str, BaseModelExample]) -> None:  # type: ignore [empty-body]
                    ...

    @parameterized.expand(
        [
            (BaseModelExample, False, TypeSpecifier(True, None, BaseModelExample, None, False, False)),
            (List[BaseModelExample], False, TypeSpecifier(True, list, BaseModelExample, None, False, False)),
            (Optional[BaseModelExample], False, TypeSpecifier(True, None, BaseModelExample, None, False, True)),
            (
                Optional[List[BaseModelExample]],
                False,
                TypeSpecifier(True, list, BaseModelExample, None, False, True),
            ),
            (List[Optional[BaseModelExample]], True, None),
            (List[BaseModelExample], False, TypeSpecifier(True, list, BaseModelExample, None, False, False)),
            (Optional[BaseModelExample], False, TypeSpecifier(True, None, BaseModelExample, None, False, True)),
            (
                Optional[List[BaseModelExample]],
                False,
                TypeSpecifier(True, list, BaseModelExample, None, False, True),
            ),
            (List[Optional[BaseModelExample]], True, None),
            (JSON, False, TypeSpecifier(True, None, None, None, True, False)),
            (str, False, TypeSpecifier(True, None, str, None, False, False)),
            (bytes, False, TypeSpecifier(True, None, bytes, None, False, False)),
            (
                Union[BaseModelExample, BaseModelExample2],
                False,
                TypeSpecifier(True, None, None, [BaseModelExample, BaseModelExample2], False, False),
            ),
            (
                Annotated[Union[BaseModelExample, BaseModelExample2], None],
                False,
                TypeSpecifier(True, None, None, [BaseModelExample, BaseModelExample2], False, False),
            ),
            (None, True, None),
        ]
    )
    def test_request_decorator_payload_spec(self, payload_type, raises, expected_payload_spec):
        # Arrange
        class TestAPI(APIEndpoints):
            def get_data(self, payload: payload_type) -> None:  # type: ignore [empty-body]
                ...

        decorator = request("POST", "/v1/data")
        # Act / Assert
        if raises:
            with self.assertRaises(APIEndpointError):
                decorator(TestAPI.get_data)
        else:
            decorator(TestAPI.get_data)
            assert decorator.payload_spec == expected_payload_spec

    def test_request_decorator_not_annotated_params(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("POST", "/v1/data")
                def get_data(self, params) -> None:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_not_supprted_param_type(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("POST", "/v1/data")
                def get_data(self, params: List[str]) -> None:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_not_annotated_url_field_arguments(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("POST", "/v1/data/{id}")
                def get_data(self, id) -> None:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_bogus_params(self):
        with self.assertRaises(APIEndpointError):

            class TestAPI(APIEndpoints):
                @request("POST", "/v1/data/{id}")  # type: ignore [empty-body]
                def get_data(self, id: str, payload: BaseModelExample, bogus: str) -> None:
                    ...

    def test_request_decorator_accepts_optional_payload(self):
        # Arrange
        class TestAPIOptional(APIEndpoints):
            @request("GET", "/v1/data")
            def get_data1(self, payload: Optional[BaseModelExample]) -> None:  # type: ignore [empty-body]
                ...

        class TestAPIUnion(APIEndpoints):
            @request("GET", "/v2/data")
            def get_data2(self, payload: Union[None, BaseModelExample]) -> None:  # type: ignore [empty-body]
                ...

        class TestAPIOptionalModelSequence(APIEndpoints):
            @request("GET", "/v3/data")
            def get_data3(self, payload: Optional[List[BaseModelExample]]) -> None:  # type: ignore [empty-body]
                ...

    def test_request_decorator_call_from_unsuitable_base_class(self):
        class TestAPI:
            @request("GET", "/v1/data")
            def get_data(self) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI()
        with self.assertRaises(APIEndpointError):
            api.get_data()

    def test_request_decorator_call_with_positional_arguments(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v1/data/{id}")
            def get_data(self, id: str, payload: BaseModelExample) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        # Act
        api.get_data("ID123", self.basemodel_payload)
        # Assert
        self.session_mock.request.assert_called_once_with(
            "GET",
            self.base_path + "/v1/data/ID123",
            data=self.basemodel_payload.model_dump_json(),
            headers={"content-type": "application/json"},
        )

    def test_request_decorator_call_with_mixed_positional_arguments(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v2/{category}/items")
            def get_data(
                self, payload: BaseModelExample, category: str, params: ParamsModelExample
            ) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        # Act
        api.get_data(self.basemodel_payload, "clothes", self.basemodel_params)
        # Assert
        self.session_mock.request.assert_called_once_with(
            "GET",
            self.base_path + "/v2/clothes/items",
            data=self.basemodel_payload.model_dump_json(),
            headers={"content-type": "application/json"},
            params=self.basemodel_params.model_dump(),
        )

    def test_request_decorator_call_with_keyword_arguments(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v2/{category}/items")
            def get_data(
                self, payload: BaseModelExample, category: str, params: ParamsModelExample
            ) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        # Act
        api.get_data(category="clothes", params=self.basemodel_params, payload=self.basemodel_payload)
        # Assert
        self.session_mock.request.assert_called_once_with(
            "GET",
            self.base_path + "/v2/clothes/items",
            data=self.basemodel_payload.model_dump_json(),
            headers={"content-type": "application/json"},
            params=self.basemodel_params.model_dump(),
        )

    def test_request_decorator_call_with_json_payload(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/aaa/bbb/ccc")
            def get_data(self, payload: JSON) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        json_payload = {"a": {"b": [1.6]}}
        # Act
        api.get_data(payload=json_payload)
        # Assert
        self.session_mock.request.assert_called_once_with(
            "GET",
            self.base_path + "/aaa/bbb/ccc",
            data=json.dumps(json_payload),
            headers={"content-type": "application/json"},
        )

    def test_request_decorator_call_optional_model_payload(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("PUT", "/v1/data/{id}")
            def put_data(self, id: str, payload: Optional[BaseModelExample]) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        # Act
        api.put_data("ID123", None)
        # Assert
        self.session_mock.request.assert_called_once_with(
            "PUT",
            self.base_path + "/v1/data/ID123",
        )

    def test_request_decorator_call_and_return_model(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v1/items")
            def get_data(self) -> BaseModelExample:  # type: ignore [empty-body]
                ...

        self.session_mock.request.return_value.dataobj = MagicMock(return_value=self.basemodel_payload)
        api = TestAPI(self.session_mock)
        # Act
        retval = api.get_data()
        # Assert
        self.session_mock.request.return_value.dataobj.assert_called_once()
        assert retval == self.basemodel_payload

    def test_request_decorator_call_and_return_model_datasequece(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v1/items")
            def get_data(self) -> DataSequence[BaseModelExample]:  # type: ignore [empty-body]
                ...

        self.session_mock.request.return_value.dataseq = MagicMock(return_value=self.basemodel_sequence_payload)
        api = TestAPI(self.session_mock)
        # Act
        retval = api.get_data()
        # Assert
        self.session_mock.request.return_value.dataseq.assert_called_once()
        assert retval == self.basemodel_sequence_payload

    def test_request_decorator_call_and_return_str(self):
        # Arrange
        expected = "This is String!"

        class TestAPI(APIEndpoints):
            @request("GET", "/v1/items")
            def get_data(self) -> str:  # type: ignore [empty-body]
                ...

        self.session_mock.request.return_value.text = expected
        api = TestAPI(self.session_mock)
        # Act
        observed = api.get_data()
        # Assert
        assert observed == expected

    def test_request_decorator_call_and_return_bytes(self):
        # Arrange
        expected = b"\xFFThis is String!"

        class TestAPI(APIEndpoints):
            @request("GET", "/v1/items")
            def get_data(self) -> bytes:  # type: ignore [empty-body]
                ...

        self.session_mock.request.return_value.content = expected
        api = TestAPI(self.session_mock)
        # Act
        observed = api.get_data()
        # Assert
        assert observed == expected

    def test_request_decorator_call_and_return_dict(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v1/items")
            def get_data(self) -> dict:  # type: ignore [empty-body]
                ...

        self.session_mock.request.return_value.json = MagicMock(return_value=self.dict_payload)
        api = TestAPI(self.session_mock)
        # Act
        retval = api.get_data()
        # Assert
        self.session_mock.request.return_value.json.assert_called_once()
        assert retval == self.dict_payload

    @parameterized.expand(
        [
            ("JStr", str),
            (1, int),
            (3.33, float),
            (True, bool),
            ({"a": {"b": [1.6]}}, dict),
            ([1, 2, {"a": {"b": [1.6]}}], list),
        ]
    )
    def test_request_decorator_call_and_return_json(self, json, jtype):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v1/items")
            def get_data(self) -> JSON:  # type: ignore [empty-body]
                ...

        self.session_mock.request.return_value.json = MagicMock(return_value=json)
        api = TestAPI(self.session_mock)
        # Act
        retval = api.get_data()
        # Assert
        self.session_mock.request.return_value.json.assert_called_once()
        assert retval == json
        assert isinstance(retval, jtype)

    def test_request_decorator_call_raises_when_payload_has_no_resp_json_key(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v1/items", "data")
            def get_data(self) -> JSON:  # type: ignore [empty-body]
                ...

        self.session_mock.request.return_value.json = MagicMock(return_value=[1, 2, 3])
        api = TestAPI(self.session_mock)
        # Act / Assert
        with self.assertRaises(TypeError):
            api.get_data()

    def test_get_decorator(self):
        # Arrange
        class TestAPI(APIEndpoints):
            def method(self) -> None:  # type: ignore [empty-body]
                ...

        decorator = get("/url/data")
        # Act / Assert
        decorator(TestAPI.method)
        decorator.http_method = "GET"

    def test_put_decorator(self):
        # Arrange
        class TestAPI(APIEndpoints):
            def method(self) -> None:  # type: ignore [empty-body]
                ...

        decorator = put("/url/data")
        # Act / Assert
        decorator(TestAPI.method)
        decorator.http_method = "PUT"

    def test_post_decorator(self):
        # Arrange
        class TestAPI(APIEndpoints):
            def method(self) -> None:  # type: ignore [empty-body]
                ...

        decorator = post("/url/data")
        # Act / Assert
        decorator(TestAPI.method)
        decorator.http_method = "POST"

    def test_delete_decorator(self):
        # Arrange
        class TestAPI(APIEndpoints):
            def method(self) -> None:  # type: ignore [empty-body]
                ...

        decorator = delete("/url/data")
        # Act / Assert
        decorator(TestAPI.method)
        decorator.http_method = "DELETE"

    def test_no_mutable_state_when_calling_endpoint(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v2/{category}/items")
            def get_data(
                self, payload: BaseModelExample, category: str, params: ParamsModelExample
            ) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        for category in ["clothes", "food"]:
            for dict_payload in [
                {"id": "id1", "size": 100, "capacity": 1.7, "active": True},
                {"id": "id2", "size": 120, "capacity": 1.9, "active": False},
            ]:
                for params in [
                    ParamsModelExample(name="submarine", color="yellow"),
                    ParamsModelExample(name="oyster", color="blue"),
                ]:
                    # Act
                    payload = BaseModelExample.model_validate(dict_payload)
                    api.get_data(category=category, params=params, payload=payload)
                    # Assert
                    self.session_mock.request.assert_called_once_with(
                        "GET",
                        self.base_path + f"/v2/{category}/items",
                        data=payload.model_dump_json(),
                        headers={"content-type": "application/json"},
                        params=params.model_dump(),
                    )
                    self.session_mock.reset_mock()

    def test_request_decorator_call_with_default_arguments(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v2/{category}/items")
            def get_data(
                self,
                payload: BaseModelExample = self.basemodel_payload,
                category: str = "default-category",
                params: ParamsModelExample = self.basemodel_params,
            ) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        # Act
        api.get_data()
        # Assert
        self.session_mock.request.assert_called_once_with(
            "GET",
            self.base_path + "/v2/default-category/items",
            data=self.basemodel_payload.model_dump_json(),
            headers={"content-type": "application/json"},
            params=self.dict_params,
        )

    def test_request_decorator_call_with_defaults_arguments_override(self):
        # Arrange
        class TestAPI(APIEndpoints):
            @request("GET", "/v2/{category}/items")
            def get_data(
                self,
                payload: BaseModelExample = self.basemodel_payload,
                category: str = "default",
                params: ParamsModelExample = self.basemodel_params,
            ) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        # Act
        payload_override = BaseModelExample(id="override-id", size=500, capacity=9.0001, active=False)
        category_override = "override-category!"
        params_override = ParamsModelExample(name="override-Name", color="override with orange!")
        api.get_data(payload_override, category_override, params_override)
        # Assert
        self.session_mock.request.assert_called_once_with(
            "GET",
            self.base_path + f"/v2/{category_override}/items",
            data=payload_override.model_dump_json(),
            headers={"content-type": "application/json"},
            params=params_override.model_dump(),
        )

    def test_decorator_chaining_order(self):
        # Expected @request can access original function signature (it will raise otherwise)
        class TestAPIMixedOrder1(APIEndpoints):
            @request("GET", "/v2/{category}/items")
            @versions("<2")
            def get_data(
                self, payload: BaseModelExample, category: str, params: ParamsModelExample
            ) -> None:  # type: ignore [empty-body]
                ...

        class TestAPIMixedOrder2(APIEndpoints):
            @request("GET", "/v2/{category}/items")
            @versions("<2")
            @view({ProviderView})
            def get_data(
                self, payload: BaseModelExample, category: str, params: ParamsModelExample
            ) -> None:  # type: ignore [empty-body]
                ...

    def test_request_decorator_format_url_with_str_enum(self):
        class FruitEnum(str, Enum):
            BANANA = "banana"
            ORANGE = "orange"
            APPLE = "apple"

        class TestAPI(APIEndpoints):
            @request("GET", "/v1/data/{fruit_type}")
            def get_data(self, fruit_type: FruitEnum, payload: str) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        # Act
        api.get_data(FruitEnum.ORANGE, "not a fruit")
        # Assert
        self.session_mock.request.assert_called_once_with("GET", self.base_path + "/v1/data/orange", data="not a fruit")

    def test_request_decorator_format_url_with_uuid(self):
        test_uuid = uuid4()

        class TestAPI(APIEndpoints):
            @request("GET", "/v1/data/{id}")
            def get_data(self, id: UUID, payload: str) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        # Act
        api.get_data(test_uuid, "not a fruit")
        # Assert
        self.session_mock.request.assert_called_once_with(
            "GET", self.base_path + f"/v1/data/{test_uuid}", data="not a fruit"
        )

    def test_request_decorator_raises_when_format_url_is_not_uuid_or_str_subtype(self):
        with self.assertRaises(APIEndpointError):

            class FruitEnum(Enum):
                BANANA = 1
                ORANGE = 2
                APPLE = 3

            class TestAPI(APIEndpoints):
                @request("POST", "/v1/data/{fruit_type}")
                def get_data(self, fruit_type: FruitEnum) -> None:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_format_url_with_literal(self):
        FruitType = Literal["banana", "orange", "apple"]

        class TestAPI(APIEndpoints):
            @request("GET", "/v1/data/{fruit_type}")
            def get_data(self, fruit_type: FruitType, payload: str) -> None:  # type: ignore [empty-body]
                ...

        api = TestAPI(self.session_mock)
        # Act
        api.get_data("banana", "not a fruit")
        # Assert
        self.session_mock.request.assert_called_once_with("GET", self.base_path + "/v1/data/banana", data="not a fruit")

    def test_request_decorator_raises_when_format_url_with_literal_is_not_str_subtype(self):
        with self.assertRaises(APIEndpointError):
            FruitType = Literal[1, 2, 3]

            class TestAPI(APIEndpoints):
                @request("POST", "/v1/data/{fruit_type}")
                def get_data(self, fruit_type: FruitType) -> None:  # type: ignore [empty-body]
                    ...

    def test_request_decorator_accept_union_of_models(self):
        class TestAPI(APIEndpoints):
            @request("GET", "/v1/data")
            def get_data(
                self, payload: Union[BaseModelExample, BaseModelExample2]
            ) -> None:  # type: ignore [empty-body]
                ...

    def test_request_decorator_accept_annotated_union_of_models(self):
        class BaseModel_A(BaseModel):
            field: Literal["number"] = "number"
            num: float

        class BaseModel_B(BaseModel):
            field: Literal["name"] = "name"
            name: str

        AnyBaseModel = Annotated[Union[BaseModel_A, BaseModel_B], Field(discriminator="field")]

        class TestAPI(APIEndpoints):
            @request("POST", "/v1/data")
            def create(self, payload: AnyBaseModel) -> None:  # type: ignore [empty-body]
                ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

import unittest
from unittest.mock import MagicMock, Mock, patch

from catalystwan.api.versions_utils import DeviceSoftwareRepository, DeviceVersions, RepositoryAPI
from catalystwan.endpoints.configuration.software_actions import SoftwareImageDetails
from catalystwan.endpoints.configuration_device_actions import InstalledDeviceData, PartitionDevice
from catalystwan.endpoints.configuration_device_inventory import DeviceDetailsResponse
from catalystwan.typed_list import DataSequence


class TestRepositoryAPI(unittest.TestCase):
    def setUp(self):
        self.device = DeviceDetailsResponse(
            personality="vedge",
            uuid="mock_uuid",
            device_ip="mock_ip",
            hostname="mock_host",
            reachability="reachable",
            local_system_ip="mock_ip",
            status="normal",
            memUsage=1.0,
            connected_vManages=["192.168.0.1"],
            model="vedge-cloud",
        )
        self.DeviceSoftwareRepository_obj = {
            "mock_uuid": DeviceSoftwareRepository(
                installed_versions=["ver1", "ver2", "curr_ver"],
                availableVersions=["ver1", "ver2"],
                version="curr_ver",
                defaultVersion="def_ver",
                uuid="mock_uuid",
            )
        }
        mock_session = Mock()
        self.mock_repository_object = RepositoryAPI(mock_session)

    @patch("catalystwan.session.Session")
    def test_get_image_version_if_image_available(self, mock_session):
        versions_response = DataSequence(
            SoftwareImageDetails,
            [SoftwareImageDetails(**{"availableFiles": "vmanage-20.9.1-x86_64.tar.gz", "versionName": "20.9.1"})],
        )
        self.mock_repository_object.get_all_software_images = MagicMock(return_value=versions_response)
        image_version = "20.9.1"
        answer = self.mock_repository_object.get_image_version("vmanage-20.9.1-x86_64.tar.gz")

        self.assertEqual(answer, image_version, "not same version")

    @patch("catalystwan.session.Session")
    def test_get_image_version_if_image_unavailable(self, mock_session):
        api_mock_response = [{"availableFiles": "vmanage-20.9.2-x86_64.tar.gz", "versionName": "20.9.1"}]
        mock_session.get_data.return_value = api_mock_response
        image_version = None
        answer = RepositoryAPI(mock_session).get_image_version("vmanage-20.9.1-x86_64.tar.gz")

        self.assertEqual(answer, image_version, "not same version")

    def test_get_remote_image(self):
        versions_response = DataSequence(
            SoftwareImageDetails,
            [
                SoftwareImageDetails(
                    **{
                        "availableFiles": "vmanage-20.9.1-x86_64.tar.gz",
                        "versionType": "remote-server-test",
                        "remoteServerId": "123456789-abcdabcd",
                        "versionId": "abcd-1234",
                    }
                )
            ],
        )
        self.mock_repository_object.get_all_software_images = MagicMock(return_value=versions_response)
        answer = self.mock_repository_object.get_remote_image("vmanage-20.9.1-x86_64.tar.gz", "remote-server-test")

        self.assertEqual(answer.version_id, "abcd-1234", "not same version")

    def test_get_remote_image_non_existing(self):
        versions_response = DataSequence(
            SoftwareImageDetails,
            [
                SoftwareImageDetails(
                    **{
                        "availableFiles": "vmanage-20.9.1-x86_64.tar.gz",
                        "versionType": "remote-server-test",
                        "remoteServerId": "123456789-abcdabcd",
                        "versionId": "abcd-1234",
                    }
                )
            ],
        )
        self.mock_repository_object.get_all_software_images = MagicMock(return_value=versions_response)
        answer = self.mock_repository_object.get_remote_image("vmanage-20.10.1-x86_64.tar.gz", "remote-server-test")

        self.assertEqual(answer, None, "not same version")

    def test_get_remote_image_no_version_id(self):
        versions_response = DataSequence(
            SoftwareImageDetails,
            [
                SoftwareImageDetails(
                    **{
                        "availableFiles": "vmanage-20.9.1-x86_64.tar.gz",
                        "versionType": "remote-server-test",
                        "remoteServerId": "123456789-abcdabcd",
                        "versionId": None,
                    }
                )
            ],
        )
        self.mock_repository_object.get_all_software_images = MagicMock(return_value=versions_response)

        with self.assertRaises(ValueError):
            self.mock_repository_object.get_remote_image("vmanage-20.9.1-x86_64.tar.gz", "remote-server-test")

    def test_get_version_when_same_available_file_present_for_remote_version(self):
        versions_response = DataSequence(
            SoftwareImageDetails,
            [
                SoftwareImageDetails(
                    **{
                        "availableFiles": "vmanage-20.9.1-x86_64.tar.gz",
                        "versionType": "remote-server-test",
                        "remoteServerId": "123456789-abcdabcd",
                        "versionId": "abcd-1234",
                    }
                ),
                SoftwareImageDetails(**{"availableFiles": "vmanage-20.9.1-x86_64.tar.gz", "versionName": "20.9.1"}),
            ],
        )
        self.mock_repository_object.get_all_software_images = MagicMock(return_value=versions_response)
        image_version = "20.9.1"
        answer = self.mock_repository_object.get_image_version("vmanage-20.9.1-x86_64.tar.gz")

        self.assertEqual(answer, image_version, "not same version")

    def test_get_version_when_remote_available_file_is_matching(self):
        versions_response = DataSequence(
            SoftwareImageDetails,
            [
                SoftwareImageDetails(
                    **{
                        "availableFiles": "vmanage-20.9.1-x86_64.tar.gz",
                        "versionType": "remote-server-test",
                        "remoteServerId": "123456789-abcdabcd",
                        "versionId": "abcd-1234",
                    }
                ),
                SoftwareImageDetails(**{"availableFiles": "vmanage-20.10.1-x86_64.tar.gz", "versionName": "20.9.1"}),
            ],
        )
        self.mock_repository_object.get_all_software_images = MagicMock(return_value=versions_response)
        answer = self.mock_repository_object.get_image_version("vmanage-20.9.1-x86_64.tar.gz")

        self.assertEqual(answer, None)

    @patch("catalystwan.session.Session")
    def test_get_devices_versions_repository(self, mock_session):
        endpoint_mock_response = DataSequence(
            InstalledDeviceData,
            [
                InstalledDeviceData(
                    **{
                        "availableVersions": ["ver1", "ver2"],
                        "version": "curr_ver",
                        "defaultVersion": "def_ver",
                        "uuid": "mock_uuid",
                    }
                )
            ],
        )

        self.mock_repository_object.session.endpoints.configuration_device_actions.get_list_of_installed_devices = (
            MagicMock(return_value=endpoint_mock_response)
        )

        answer = self.mock_repository_object.get_devices_versions_repository()

        self.assertEqual(
            answer["mock_uuid"],
            self.DeviceSoftwareRepository_obj["mock_uuid"],
            "DeviceSoftwareRepository object created uncorrectly",
        )

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_get_device_available(self, mock_get_devices_versions_repository):
        # Prepare mock data
        mock_get_devices_versions_repository.return_value = Mock()
        mock_session = Mock()
        mock_repository_object = RepositoryAPI(mock_session)
        mock_device_versions = DeviceVersions(mock_repository_object)
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        answer = mock_device_versions.get_device_available("ver1", [self.device])
        expected_result = DataSequence(
            PartitionDevice, [PartitionDevice(device_id="mock_uuid", device_ip="mock_ip", version="ver1")]
        )

        # Assert
        self.assertEqual(
            answer,
            expected_result,
            "Version add incorrectly",
        )

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_get_device_list_if_in_installed(self, mock_get_devices_versions_repository):
        # Prepare mock data
        mock_get_devices_versions_repository.return_value = Mock()
        mock_session = Mock()
        mock_repository_object = RepositoryAPI(mock_session)
        mock_device_versions = DeviceVersions(mock_repository_object)
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        answer = mock_device_versions.get_device_list_in_installed("ver1", [self.device])
        expected_result = DataSequence(
            PartitionDevice, [PartitionDevice(device_id="mock_uuid", device_ip="mock_ip", version="ver1")]
        )

        # Assert
        self.assertEqual(
            answer,
            expected_result,
            "Version add incorrectly",
        )

    @patch.object(RepositoryAPI, "get_devices_versions_repository")
    def test_get_devices_current_version(self, mock_get_devices_versions_repository):
        # Arrange
        mock_session = Mock()
        mock_get_devices_versions_repository.return_value = Mock()
        mock_get_devices_versions_repository.return_value = self.DeviceSoftwareRepository_obj
        mock_repository_object = RepositoryAPI(mock_session)
        mock_device_versions = DeviceVersions(mock_repository_object)
        # Act
        answer = mock_device_versions.get_devices_current_version([self.device])
        # Answer
        proper_answer = DataSequence(
            PartitionDevice, [PartitionDevice(device_id="mock_uuid", device_ip="mock_ip", version="curr_ver")]
        )
        self.assertEqual(answer, proper_answer)



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from typing import List, Literal, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable

AttachmentType = Literal[
    "custom",
    "auto",
]

TrafficDirection = Literal[
    "UNIDIR",
    "BIDIR",
]

ReachableInterfaceType = Literal[
    "ipv4",
    "ipv6",
    "tunnel",
]

TrackType = Literal[
    "service-icmp",
    "ipv6-service-icmp",
]

InterfaceType = Literal[
    "fromservice",
    "toservice",
]

RedundancyType = Literal[
    "active",
    "backup",
]

ServiceType = Literal[
    "Firewall",
    "Intrusion-detection",
    "Intrusion-detection-prevention",
    "NETSVC1",
    "NETSVC2",
    "NETSVC3",
    "NETSVC4",
    "NETSVC5",
    "NETSVC6",
    "NETSVC7",
]


class GatewayInterface(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="gatewayInterfaceName", validation_alias="gatewayInterfaceName", default=None
    )
    ip_address: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="gatewayIpAddress", validation_alias="gatewayIpAddress", default=None
    )
    ipv6_address: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="gatewayIpv6Address", validation_alias="gatewayIpv6Address", default=None
    )


class ServiceInterface(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_name: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="serviceInterfaceName", validation_alias="serviceInterfaceName", default=None
    )
    ip_address: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="serviceIpAddress", validation_alias="serviceIpAddress", default=None
    )
    ipv6_address: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="serviceIpv6Address", validation_alias="serviceIpv6Address", default=None
    )


class TrackingIP(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    ipv4: Optional[Union[Global[str], Variable]] = None
    ipv6: Optional[Union[Global[str], Variable]] = None


class ReachableInterface(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    type: Global[ReachableInterfaceType] = Field(
        serialization_alias="reachableInterfaceType", validation_alias="reachableInterfaceType"
    )
    name: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="reachableInterfaceName", validation_alias="reachableInterfaceName"
    )
    ip_address: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="reachableIpAddress", validation_alias="reachableIpAddress", default=None
    )
    ipv6_address: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="reachableIpv6Address", validation_alias="reachableIpv6Address", default=None
    )
    tracking_ip: Optional[TrackingIP] = Field(
        serialization_alias="trackingIp", validation_alias="trackingIp", default=None
    )
    track_name: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="trackName", validation_alias="trackName", default=None
    )
    track_type: Optional[Union[Global[TrackType], Variable]] = Field(
        serialization_alias="trackType", validation_alias="trackType", default=None
    )


class InterfaceProperties(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    gateway_interface: Optional[GatewayInterface] = Field(
        serialization_alias="gatewayInterface", validation_alias="gatewayInterface", default=None
    )
    service_interface: Optional[ServiceInterface] = Field(
        serialization_alias="serviceInterface", validation_alias="serviceInterface", default=None
    )
    reachable_interface: Optional[ReachableInterface] = Field(
        serialization_alias="reachableInterface", validation_alias="reachableInterface", default=None
    )
    interface_type: Optional[Union[Global[InterfaceType], Variable]] = Field(
        serialization_alias="interfaceType", validation_alias="interfaceType", default=None
    )
    redundancy_type: Optional[Union[Global[RedundancyType], Variable]] = Field(
        serialization_alias="redundancyType", validation_alias="redundancyType", default=None
    )


class Attachment(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interface_properties: List[InterfaceProperties] = Field(
        serialization_alias="interfaceProperties", validation_alias="interfaceProperties"
    )
    traffic_direction: Optional[Union[Global[TrafficDirection], Variable]] = Field(
        serialization_alias="trafficDirection", validation_alias="trafficDirection", default=None
    )


class TrackConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    interval: Optional[Union[Global[int], Variable]] = None
    threshold: Optional[Union[Global[int], Variable]] = None
    multiplier: Optional[Union[Global[int], Variable]] = None


class Service(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    service_type: Union[Global[ServiceType], Variable] = Field(
        serialization_alias="serviceType", validation_alias="serviceType"
    )
    default_track: Optional[Union[Global[bool], Default[bool]]] = Field(
        serialization_alias="defaultTrack", validation_alias="defaultTrack", default=None
    )
    track: Optional[Global[bool]] = None
    track_config: Optional[TrackConfig] = Field(
        serialization_alias="trackConfig", validation_alias="trackConfig", default=None
    )
    attachments: Optional[List[Attachment]] = None


class ServiceInsertionAttachmentData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    attachment_type: Optional[Union[Global[AttachmentType], Variable]] = Field(
        serialization_alias="attachmentType", validation_alias="attachmentType", default=None
    )
    service_chain_instance_id: Optional[Union[Global[str], Variable]] = Field(
        serialization_alias="serviceChainInstanceID", validation_alias="serviceChainInstanceID", default=None
    )
    service_chain_definition_id: Global[UUID] = Field(
        serialization_alias="serviceChainDefinitionID", validation_alias="serviceChainDefinitionID"
    )
    vpn: Union[Global[int], Variable]
    services: Optional[List[Service]] = None


class ServiceInsertionAttachmentCreationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    name: str
    description: Optional[str] = None
    data: ServiceInsertionAttachmentData



# Copyright 2024 Cisco Systems, Inc. and its affiliates

from ipaddress import IPv6Address, IPv6Network
from typing import List

from pydantic import AliasPath, BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Global, _ParcelBase, as_global


class IPv6PrefixListEntry(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    ipv6_address: Global[IPv6Address] = Field(serialization_alias="ipv6Address", validation_alias="ipv6Address")
    ipv6_prefix_length: Global[int] = Field(serialization_alias="ipv6PrefixLength", validation_alias="ipv6PrefixLength")


class IPv6PrefixListParcel(_ParcelBase):
    entries: List[IPv6PrefixListEntry] = Field(default=[], validation_alias=AliasPath("data", "entries"))

    def add_prefix(self, ipv6_network: IPv6Network):
        self.entries.append(
            IPv6PrefixListEntry(
                ipv6_address=as_global(ipv6_network.network_address),
                ipv6_prefix_length=as_global(ipv6_network.prefixlen),
            )
        )



# Copyright 2024 Cisco Systems, Inc. and its affiliates

# type: ignore
import copy
import unittest
from unittest import TestCase

from attr import define  # type: ignore
from parameterized import parameterized

from catalystwan.dataclasses import DataclassBase, Device, User
from catalystwan.exceptions import InvalidOperationError
from catalystwan.typed_list import DataSequence, TypedList


@define
class FakeUser(DataclassBase):
    name: str
    weight: float


class TestTypedList(TestCase):
    def setUp(self):
        self.u = User(username="User1")
        u1 = User(username="User1")
        u2 = User(username="User2")
        u3 = User(username="User3")
        self.users = [u1, u2, u3]
        self.typed_list = TypedList(User, self.users)

    @parameterized.expand([(str, []), (int, [1, 2, 3]), (User, [User("User1")])])
    def test_init(self, _type, iterable):
        # Arrange, Act
        TypedList(_type, iterable)

    @parameterized.expand([(str, [1]), (int, [1, 2, "3"]), (User, [User(username="User1"), 1])])  # type: ignore
    def test_init_type_error(self, _type, iterable):
        # Arrange, Act, Assert
        with self.assertRaises(TypeError):
            TypedList(_type, iterable)

    def test_init_no_args(self):
        # Arrange, Act, Assert
        with self.assertRaises(TypeError):
            TypedList()

    def test_init_only_type(self):
        # Arrange, Act, Assert
        TypedList(str)

    @parameterized.expand([(str, [], "TypedList(str, [])"), (int, [1, 2, 3], "TypedList(int, [1, 2, 3])")])
    def test_repr_trival(self, _type, iterable, output):
        # Arrange, Act
        representation = repr(TypedList(_type, iterable))

        # Assert
        self.assertEqual(representation, output)

    # Integration
    def test_repr(self):
        # Arrange
        typed_list = TypedList(User, self.users)
        output = (
            "TypedList(User, ["
            "User(username='User1', password=None, group=[], locale=None, description=None, resource_group=None), "
            "User(username='User2', password=None, group=[], locale=None, description=None, resource_group=None), "
            "User(username='User3', password=None, group=[], locale=None, description=None, resource_group=None)])"
        )
        # Act
        representation = repr(typed_list)

        # Assert
        self.assertEqual(representation, output)

    @parameterized.expand(
        [
            (str, ["1"], 1),
            (int, [1, 2, 3], 3),
            (User, None, 0),
            (User, [User(username="User1")], 1),
            (User, [User(username="User1"), User(username="User2")], 2),
        ]
    )
    def test_len(self, _type, iterable, length):
        # Arrange, Act
        output_length = len(TypedList(_type, iterable))

        # Assert
        self.assertEqual(output_length, length)

    @parameterized.expand(
        [
            (str, ["1"], "1"),
            (int, [1, 2, 3], 3),
            (User, [User(username="User1")], User(username="User1")),
            (User, [User(username="User1"), User(username="User2")], User(username="User1")),
        ]
    )
    def test_contains_positive(self, _type, iterable, other):
        # Arrange, Act
        typed_list = TypedList(_type, iterable)

        # Assert
        self.assertTrue(other in typed_list)

    @parameterized.expand(
        [
            (str, ["1"], "2"),
            (int, [1, 2, 3], 4),
            (User, [User(username="User1")], User(username="User2")),
            (User, [User(username="User1"), User(username="User2")], User(username="User3")),
        ]
    )
    def test_negative(self, _type, iterable, other):
        # Arrange, Act
        typed_list = TypedList(_type, iterable)

        # Assert
        self.assertFalse(other in typed_list)

    @parameterized.expand(
        [
            (1, User(username="User2")),
            (slice(1, 3, 1), TypedList(User, [User(username="User2"), User(username="User3")])),
        ]
    )
    def test_get_item(self, index, output):
        self.assertEqual(self.typed_list[index], output)

    @parameterized.expand([(User(username="User1"),), (TypedList(User, [User("User3")]),)])
    def test_eq_negative(self, other):
        self.assertFalse(self.users == other)

    def test_eq_false_sort(self):
        # Arrange
        u1 = User(username="User1")
        u2 = User(username="User2")
        u3 = User(username="User3")

        # Act, Assert
        self.assertFalse(self.users == [u1, u3, u2])

    def test_eq(self):
        # Arrange
        u1 = User(username="User1")
        u2 = User(username="User2")
        u3 = User(username="User3")

        # Act, Assert
        self.assertTrue(self.users == [u1, u2, u3])

    def test_set_item(self):
        # Arrange
        u1 = User(username="User1")
        u2 = User(username="User2")
        u3 = User(username="User3")

        users = TypedList(User, [u1, u2])

        # Act
        users[1] = u3

        # Assert
        self.assertEqual(users[1], u3)
        self.assertEqual(users[0], u1)
        with self.assertRaises(IndexError):
            users[2]
        self.assertNotEqual(users[1], u2)
        with self.assertRaises(TypeError):
            users[1] = 2

    def test_insert(self):
        # Arrange
        u1 = User(username="User1")
        u2 = User(username="User2")
        u3 = User(username="User3")

        users = TypedList(User, [u1, u2])

        # Act
        users.insert(1, u3)

        # Assert
        self.assertEqual(users[0], u1)
        self.assertEqual(users[1], u3)
        self.assertEqual(users[2], u2)
        with self.assertRaises(IndexError):
            users[3]
        self.assertNotEqual(users[1], u2)
        with self.assertRaises(TypeError):
            users.insert(1, "2")

    @parameterized.expand(
        [
            (TypedList(str, ["A", "B"]), TypedList(str, ["C"]), False, TypedList(str, ["A", "B", "C"])),
            (TypedList(int, [1, 2]), TypedList(int, [5]), False, TypedList(int, [1, 2, 5])),
            (TypedList(str, ["Z"]), TypedList(int, [9]), True, None),
        ]
    )
    def test_add(self, term1, term2, raises, expected_result):
        if raises:
            with self.assertRaises(TypeError):
                observed_result = term1 + term2
        else:
            observed_result = term1 + term2
            assert observed_result == expected_result

    @parameterized.expand(
        [
            (TypedList(str, ["A", "B"]), TypedList(str, ["C"]), False, TypedList(str, ["A", "B", "C"])),
            (TypedList(int, [1, 2]), TypedList(int, [5]), False, TypedList(int, [1, 2, 5])),
            (TypedList(str, ["Z"]), TypedList(int, [9]), True, None),
        ]
    )
    def test_iadd(self, term1, term2, raises, expected_result):
        if raises:
            with self.assertRaises(TypeError):
                term1 += term2
        else:
            term1 += term2
            assert term1 == expected_result


class TestDataSequence(TestCase):
    def setUp(self):
        self.u = User(username="User1")
        u1 = User(username="User1")
        u2 = User(username="User2")
        u3 = User(username="User3")
        self.users = [u1, u2, u3]
        self.data_sequence = DataSequence(User, self.users)

    @parameterized.expand(
        [
            (User, [], []),
            (User, None, []),
            (User, [User(username="User1")], [User(username="User1")]),
            (Device, [], []),
        ]
    )
    def test_init(self, _type, items, data):
        # Arrange, Act
        data_seq = DataSequence(_type, items)

        # Assert
        self.assertEqual(data_seq.data, data)

    def test_init_exception(self):
        # Arrage, Act, Assert
        with self.assertRaises(TypeError):
            DataSequence(str, ["1, 2"])

    @parameterized.expand(
        [
            (
                DataSequence(User, [User(username="A")]),
                DataSequence(User, [User(username="B")]),
                False,
                DataSequence(User, [User(username="A"), User(username="B")]),
            ),
            (
                DataSequence(User, [User(username="A")]),
                DataSequence(FakeUser, [FakeUser(name="B", weight=99.9)]),
                True,
                None,
            ),
        ]
    )
    def test_add(self, term1, term2, raises, expected_result):
        if raises:
            with self.assertRaises(TypeError):
                observed_result = term1 + term2
        else:
            observed_result = term1 + term2
            assert observed_result == expected_result

    @parameterized.expand(
        [
            (
                DataSequence(User, [User(username="A")]),
                DataSequence(User, [User(username="B")]),
                False,
                DataSequence(User, [User(username="A"), User(username="B")]),
            ),
            (
                DataSequence(User, [User(username="A")]),
                DataSequence(FakeUser, [FakeUser(name="B", weight=99.9)]),
                True,
                None,
            ),
        ]
    )
    def test_iadd(self, term1, term2, raises, expected_result):
        if raises:
            with self.assertRaises(TypeError):
                term1 += term2
        else:
            term1 += term2
            assert term1 == expected_result

    def test_single_or_default(self):
        # Arrange, Act
        output = DataSequence(User, [self.u]).single_or_default(default="x")

        # Assert
        self.assertEqual(output, self.u)

    @parameterized.expand(
        [
            (User, [], None),
            (User, [], "Test"),
        ]
    )
    def test_single_or_default_empty(self, _type, items, default_value):
        # Arrange
        data_seq = DataSequence(_type, items)

        # Act
        output = data_seq.single_or_default(default=default_value)

        # Assert
        self.assertEqual(output, default_value)

    def test_single_or_default_big_array(self):
        # Arrange
        data_seq = DataSequence(User, self.users)

        # Act, Assert
        with self.assertRaises(InvalidOperationError):
            data_seq.single_or_default()

    @parameterized.expand(
        [
            (None, []),
            ("DoesNotExists", []),
            ("User1", [User("User1")]),
        ]
    )
    def test_filter_single_attribute(self, attribute, users):
        # Arrange
        correct_seq = DataSequence(User, users)

        # Act
        output = self.data_sequence.filter(username=attribute)

        # Assert
        self.assertEqual(output, correct_seq)
        self.assertTrue(isinstance(output, DataSequence))

    def test_filter_two_attributes(self):
        # Arrange
        additional_user = User(username="User1", description="ThisOne")
        users = copy.deepcopy(self.data_sequence)
        users.append(additional_user)
        correct_output = DataSequence(User, [additional_user])

        # Act
        output = users.filter(username="User1", description="ThisOne")

        # Assert
        self.assertEqual(output, correct_output)

    def test_filter_exception(self):
        # Arrange, Act, Assert
        with self.assertRaises(AttributeError):
            self.data_sequence.filter(does_not="exists")

    def test_find(self):
        # Arrange
        expected = self.data_sequence.filter(username="User3").single_or_default()
        # Act
        observed = self.data_sequence.find(username="User3")

        # Assert
        self.assertEqual(observed, expected)

    def test_find_two_attributes(self):
        # Arrange
        additional_user = User(username="User1", description="ThisOne")
        users = copy.deepcopy(self.data_sequence)
        users.append(additional_user)
        correct_output = additional_user

        # Act
        output = users.find(username="User1", description="ThisOne")

        # Assert
        self.assertEqual(output, correct_output)

    def test_find_wrong_arg(self):
        # Arrange, Act, Assert
        with self.assertRaises(AttributeError):
            self.data_sequence.find(does_not="exists")

    def test_find_no_match(self):
        # Arrange, Act, Assert
        with self.assertRaises(InvalidOperationError):
            self.data_sequence.find(username="NonExistingUser")

    def test_deepcopy_basemodel(self):
        # Arrange
        user = User(username="admin")
        ds = DataSequence(User, [user])
        # Act
        ds_deepcopy = copy.deepcopy(ds)
        user_deepcopied = ds_deepcopy[0]
        # Assert
        self.assertIsNot(ds, ds_deepcopy)
        self.assertIsNot(user, user_deepcopied)
        self.assertEqual(user.username, user_deepcopied.username)

    def test_deepcopy_dataclass(self):
        # Arrange
        fu = FakeUser(name="admin", weight=87.5)
        ds = DataSequence(FakeUser, [fu])
        # Act
        ds_deepcopy = copy.deepcopy(ds)
        fd_deepcopied = ds_deepcopy[0]
        # Assert
        self.assertIsNot(ds, ds_deepcopy)
        self.assertIsNot(fu, fd_deepcopied)
        self.assertEqual(fu.name, fd_deepcopied.name)


if __name__ == "__main__":
    unittest.main()



# Copyright 2023 Cisco Systems, Inc. and its affiliates

from datetime import datetime
from typing import Generic, List, Literal, Optional, TypeVar, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.api.configuration_groups.parcel import Default, Global, Variable, as_global
from catalystwan.models.configuration.common import Solution

T = TypeVar("T")


IPV4Address = str
IPv6Address = str

ParcelType = Literal[
    "appqoe",
    "lan/vpn",
    "lan/vpn/interface/ethernet",
    "lan/vpn/interface/gre",
    "lan/vpn/interface/ipsec",
    "lan/vpn/interface/svi",
    "dhcp-server",
    "tracker",
    "trackergroup",
    "routing/bgp",
    "routing/eigrp",
    "routing/multicast",
    "routing/ospf",
    "routing/ospfv3/ipv4",
    "routing/ospfv3/ipv6",
    "wirelesslan",
    "switchport",
    "app-probe",
    "app-list",
    "color",
    "data-prefix",
    "expanded-community",
    "class",
    "data-ipv6-prefix",
    "ipv6-prefix",
    "prefix",
    "policer",
    "preferred-color-group",
    "sla-class",
    "tloc",
    "standard-community",
    "security-localdomain",
    "security-fqdn",
    "security-ipssignature",
    "security-urllist",
    "security-urllist",
    "security-port",
    "security-protocolname",
    "security-geolocation",
    "security-zone",
    "security-localapp",
    "security-data-ip-prefix",
]

ProfileType = Literal[
    "transport",
    "system",
    "cli",
    "service",
    "application-priority",
    "policy-object",
    "embedded-security",
]

SchemaType = Literal[
    "post",
    "put",
]


class FeatureProfileInfo(BaseModel):
    profile_id: UUID = Field(alias="profileId")
    profile_name: str = Field(alias="profileName")
    solution: Solution
    profile_type: ProfileType = Field(alias="profileType")
    created_by: str = Field(alias="createdBy")
    last_updated_by: str = Field(alias="lastUpdatedBy")
    description: str
    created_on: datetime = Field(alias="createdOn")
    last_updated_on: datetime = Field(alias="lastUpdatedOn")


class FeatureProfileDetail(BaseModel):
    profile_id: str = Field(alias="profileId")
    profile_name: str = Field(alias="profileName")
    solution: Solution
    profile_type: ProfileType = Field(alias="profileType")
    created_by: str = Field(alias="createdBy")
    last_updated_by: str = Field(alias="lastUpdatedBy")
    description: str
    created_on: datetime = Field(alias="createdOn")
    last_updated_on: datetime = Field(alias="lastUpdatedOn")
    associated_profile_parcels: List[str] = Field(alias="associatedProfileParcels")
    rid: int = Field(alias="@rid")
    profile_parcel_count: int = Field(alias="profileParcelCount")
    cached_profile: Optional[str] = Field(alias="cachedProfile")


class FromFeatureProfile(BaseModel):
    copy_: UUID = Field(alias="copy")


class FeatureProfileCreationPayload(BaseModel):
    name: str
    description: str
    from_feature_profile: Optional[FromFeatureProfile] = Field(alias="fromFeatureProfile", default=None)


class FeatureProfileEditPayload(BaseModel):
    name: str
    description: str


class FeatureProfileCreationResponse(BaseModel):
    id: UUID


class ParcelCreationResponse(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    id: UUID = Field(serialization_alias="parcelId", validation_alias="parcelId")


class Parcel(BaseModel, Generic[T]):
    parcel_id: str = Field(alias="parcelId")
    parcel_type: ParcelType = Field(alias="parcelType")
    created_by: str = Field(alias="createdBy")
    last_updated_by: str = Field(alias="lastUpdatedBy")
    created_on: int = Field(alias="createdOn")
    last_updated_on: int = Field(alias="lastUpdatedOn")
    payload: T


class Header(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    generated_on: int = Field(alias="generatedOn")


class ParcelInfo(BaseModel, Generic[T]):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    header: Header
    data: List[Parcel[T]]


class ParcelAssociationPayload(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)

    parcel_id: str = Field(alias="parcelId")


class Prefix(BaseModel):
    address: Union[Variable, Global[str]]
    mask: Union[Variable, Global[str]]


class SchemaTypeQuery(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    schema_type: SchemaType = Field(alias="schemaType")


class ParcelId(BaseModel):
    id: str = Field(alias="parcelId")


class GetFeatureProfilesPayload(BaseModel):
    limit: Optional[int]
    offset: Optional[int]


class ParcelSequence(BaseModel, Generic[T]):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    header: Header
    data: List[Parcel[T]]


class DNSIPv4(BaseModel):
    primary_dns_address_ipv4: Union[Default[None], Global[str], Variable] = Field(
        default=Default[None](value=None), alias="primaryDnsAddressIpv4"
    )
    secondary_dns_address_ipv4: Union[Default[None], Global[str], Variable] = Field(
        default=Default[None](value=None), alias="secondaryDnsAddressIpv4"
    )


class DNSIPv6(BaseModel):
    primary_dns_address_ipv6: Union[Default[None], Global[str], Variable] = Field(
        default=Default[None](value=None), alias="primaryDnsAddressIpv6"
    )
    secondary_dns_address_ipv6: Union[Default[None], Global[str], Variable] = Field(
        default=Default[None](value=None), alias="secondaryDnsAddressIpv6"
    )


class HostMapping(BaseModel):
    host_name: Union[Global[str], Variable] = Field(alias="hostName")
    list_of_ips: Union[Global[List[str]], Variable] = Field(alias="listOfIp")


class NextHop(BaseModel):
    address: Union[Global[str], Variable] = Field()
    distance: Union[Default[int], Global[int], Default[int]] = Field(default=Default[int](value=1))


class IPv4Prefix(BaseModel):
    ip_address: Union[Global[IPV4Address], Variable] = Field()
    subnet_mask: Union[Global[str], Variable] = Field()


class WANIPv4StaticRoute(BaseModel):
    prefix: IPv4Prefix = Field()
    gateway: Global[Literal["nextHop", "null0", "dhcp"]] = Field(default=Global(value="nextHop"), alias="gateway")
    next_hops: Optional[List[NextHop]] = Field(default_factory=list, alias="nextHop")
    distance: Optional[Global[int]] = Field(default=None, alias="distance")

    def set_to_next_hop(
        self,
        prefix: Optional[Global[str]] = None,
        next_hops: Optional[List[NextHop]] = None,
    ):
        if prefix is not None:
            self.prefix = as_global(prefix)
        self.gateway = Global[Literal["nextHop", "null0", "dhcp"]](value="nextHop")
        self.next_hops = next_hops or []
        self.distance = None

    def set_to_null0(
        self,
        prefix: Optional[IPv4Prefix] = None,
        distance: Union[Global[int], int] = 1,
    ):
        if prefix is not None:
            self.prefix = prefix
        self.gateway = Global(value="null0")
        self.next_hops = None
        if isinstance(distance, int):
            self.distance = Global(value=distance)
        else:
            self.distance = distance

    def set_to_dhcp(
        self,
        prefix: Optional[IPv4Prefix] = None,
    ):
        if prefix is not None:
            self.prefix = prefix
        self.gateway = Global(value="dhcp")
        self.next_hops = None
        self.distance = None


class NextHopContainer(BaseModel):
    next_hop: List[NextHop] = Field(default=[], alias="nextHop")


class Ipv6StaticRouteNull0(BaseModel):
    null0: Union[Default[bool], Global[bool]] = Field(default=Default[bool](value=True))


class IPv6StaticRouteNextHop(BaseModel):
    next_hop_container: Optional[NextHopContainer] = Field(default=None)


class IPv6StaticRouteNAT(BaseModel):
    nat: Union[Variable, Global[Literal["NAT64", "NAT66"]]] = Field()


class WANIPv6StaticRoute(BaseModel):
    prefix: Global[IPv6Address] = Field()
    gateway: Union[Ipv6StaticRouteNull0, IPv6StaticRouteNextHop, IPv6StaticRouteNAT] = Field(alias="oneOfIpRoute")

    def set_to_next_hop(
        self,
        prefix: Optional[IPv6Address] = None,
        next_hops: Optional[List[NextHop]] = None,
    ):
        if prefix is not None:
            self.prefix = as_global(prefix)
        if next_hops:
            self.gateway = IPv6StaticRouteNextHop(next_hop_container=NextHopContainer(nextHop=next_hops))

    def set_to_null0(
        self,
        prefix: Optional[IPv6Address] = None,
        enabled: Union[Default[bool], Global[bool], None] = None,
    ):
        if prefix is not None:
            self.prefix = as_global(prefix)
        if enabled is None:
            enabled = Default[bool](value=True)
        self.gateway = Ipv6StaticRouteNull0(null0=enabled)

    def set_to_nat(
        self,
        prefix: Optional[IPv6Address],
        nat: Union[Variable, Global[Literal["NAT64", "NAT66"]]],
    ):
        if prefix is not None:
            self.prefix = as_global(prefix)
        self.gateway = IPv6StaticRouteNAT(nat=nat)


class WANService(BaseModel):
    service_type: Global[Literal["TE"]] = Field(alias="serviceType")



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# mypy: disable-error-code="empty-body"

from enum import Enum
from typing import List, Optional, Union
from uuid import UUID

from pydantic import BaseModel, ConfigDict, Field

from catalystwan.endpoints import APIEndpoints, delete, post
from catalystwan.typed_list import DataSequence


class DeviceDeletionResponse(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    local_delete_from_db: Optional[bool] = Field(default=None, alias="localDeleteFromDB")
    id: Optional[UUID] = Field(default=None)


class TargetDevice(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    device_ip: str = Field(alias="deviceIP")


class DeviceCsrGenerationResponse(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    globalState: Optional[str] = Field(default=None, alias="globalState")
    discoveredDeviceVpns: Optional[List[str]] = Field(default=None, alias="discoveredDeviceVpns")
    deviceConfigurationRfs: Optional[str] = Field(default=None, alias="deviceConfigurationRfs")
    configStatusMessageDetails: Optional[str] = Field(default=None, alias="configStatusMessageDetails")
    isPrestagingSupported: Optional[bool] = Field(default=None, alias="isPrestagingSupported")
    expirationStatus: Optional[str] = Field(default=None, alias="expirationStatus")
    uuid: Optional[UUID] = Field(default=None, alias="uuid")
    certificateDetail: Optional[str] = Field(default=None, alias="certificateDetail")
    createdAt: Optional[str] = Field(default=None, alias="createdAt")
    password: Optional[str] = Field(default=None, alias="password")
    solution: Optional[str] = Field(default=None, alias="solution")
    configOperationMode: Optional[str] = Field(default=None, alias="configOperationMode")
    state: Optional[str] = Field(default=None, alias="state")
    deviceType: Optional[str] = Field(default=None, alias="deviceType")
    serialNumber: Optional[str] = Field(default=None, alias="serialNumber")
    rootCertHash: Optional[str] = Field(default=None, alias="rootCertHash")
    deviceCSRCommonName: Optional[str] = Field(default=None, alias="deviceCSRCommonName")
    deviceConfiguration: Optional[str] = Field(default=None, alias="deviceConfiguration")
    CSRDetail: Optional[str] = Field(default=None, alias="CSRDetail")
    managementSystemIP: Optional[str] = Field(default=None, alias="managementSystemIP")
    deviceCSRGenTime: Optional[int] = Field(default=None, alias="deviceCSRGenTime")
    volatileUUID: Optional[UUID] = Field(default=None, alias="volatileUUID")
    deviceLifeCycleNeeded: Optional[bool] = Field(default=None, alias="deviceLifeCycleNeeded")
    deviceState: Optional[str] = Field(default=None, alias="deviceState")
    expirationDateLong: Optional[int] = Field(default=None, alias="expirationDateLong")
    timeRemainingForExpiration: Optional[int] = Field(default=None, alias="timeRemainingForExpiration")
    certInstallStatus: Optional[str] = Field(default=None, alias="certInstallStatus")
    activity: Optional[List[str]] = Field(default=None, alias="activity")
    deviceCSR: Optional[str] = Field(default=None, alias="deviceCSR")
    discoveredDeviceInterfaces: Optional[List[str]] = Field(default=None, alias="discoveredDeviceInterfaces")
    configuredSystemIP: Optional[str] = Field(default=None, alias="configuredSystemIP")
    platformFamily: Optional[str] = Field(default=None, alias="platformFamily")
    valid: Optional[str] = Field(default=None, alias="valid")
    templateStatus: Optional[str] = Field(default=None, alias="templateStatus")
    rid: Optional[int] = Field(default=None, alias="@rid")
    personality: Optional[str] = Field(default=None, alias="personality")
    hardwareVedge: Optional[bool] = Field(default=None, alias="hardwareVedge")
    chasisNumber: Optional[str] = Field(default=None, alias="chasisNumber")
    expirationDate: Optional[str] = Field(default=None, alias="expirationDate")
    deviceIP: Optional[str] = Field(default=None, alias="deviceIP")
    uploadSource: Optional[str] = Field(default=None, alias="uploadSource")
    deviceEnterpriseCertInstallTime: Union[str, int, None] = Field(
        default=None, alias="deviceEnterpriseCertInstallTime"
    )
    CSR: Optional[str] = Field(default=None, alias="CSR")
    deviceCSRGenTimeString: Optional[str] = Field(default=None, alias="deviceCSRGenTimeString")
    deviceEnterpriseCertificate: Optional[str] = Field(default=None, alias="deviceEnterpriseCertificate")
    ncsDeviceName: Optional[str] = Field(default=None, alias="ncsDeviceName")
    configStatusMessage: Optional[str] = Field(default=None, alias="configStatusMessage")
    requestTokenID: Optional[str] = Field(default=None, alias="requestTokenID")
    templateApplyLog: Optional[List[str]] = Field(default=None, alias="templateApplyLog")
    host_name: Optional[str] = Field(default=None, alias="host-name")
    createdBy: Optional[str] = Field(default=None, alias="createdBy")
    deviceModel: Optional[str] = Field(default=None, alias="deviceModel")
    validity: Optional[str] = Field(default=None, alias="validity")
    servicePersonality: Optional[str] = Field(default=None, alias="servicePersonality")
    configuredSiteId: Optional[str] = Field(default=None, alias="configuredSiteId")
    username: Optional[str] = Field(default=None, alias="username")
    local_system_ip: Optional[str] = Field(default=None, alias="local-system-ip")
    system_ip: Optional[str] = Field(default=None, alias="system-ip")
    site_id: Optional[str] = Field(default=None, alias="site-id")


class Validity(str, Enum):
    VALID = "valid"
    INVALID = "invalid"


class VedgeListValidityPayload(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    chasis_number: str = Field(alias="chasisNumber")
    serial_number: Optional[str] = Field(default=None, alias="serialNumber")
    validity: Validity = Field(default=Validity.INVALID)


class CertActionResponse(BaseModel):
    id: UUID


class CertificateManagementDevice(APIEndpoints):
    @delete("/certificate/{uuid}")
    def delete_configuration(self, uuid: str) -> DeviceDeletionResponse:
        ...

    @post("/certificate/generate/csr", "data")
    def generate_csr(self, payload: TargetDevice) -> DataSequence[DeviceCsrGenerationResponse]:
        ...

    @post("/certificate/save/vedge/list")
    def change_vedge_list_validity(self, payload: List[VedgeListValidityPayload]) -> CertActionResponse:
        ...

    @post("/certificate/vedge/list?action={action}")
    def send_to_controllers(self, action: str = "push") -> CertActionResponse:
        ...

    @post("/certificate/vsmart/list")
    def send_to_vbond(self) -> CertActionResponse:
        ...



# Copyright 2023 Cisco Systems, Inc. and its affiliates

# type: ignore

from catalystwan.api.templates.models.cisco_vpn_model import (
    Advertise,
    AdvertiseProtocol,
    AdvertiseProtocolSubType,
    CiscoVPNModel,
    Direction,
    Dns,
    DnsIpv6,
    GreRoute,
    Host,
    IpsecRoute,
    Ipv6Advertise,
    Ipv6AdvertiseProtocol,
    Ipv6AdvertiseProtocolSubType,
    LeakFromGlobalProtocol,
    Nat,
    Natpool,
    NextHop,
    NextHopv6,
    Overload,
    Pool,
    PortForward,
    PrefixList,
    Proto,
    Region,
    Role,
    RouteExport,
    RouteExportProtocol,
    RouteExportProtocolSubType,
    RouteExportRedistribute,
    RouteExportRedistributeProtocol,
    RouteImport,
    RouteImportFrom,
    RouteImportFromProtocol,
    RouteImportFromProtocolSubType,
    RouteImportFromRedistribute,
    RouteImportFromRedistributeProtocol,
    RouteImportProtocol,
    RouteImportProtocolSubType,
    RouteImportRedistribute,
    RouteImportRedistributeProtocol,
    Routev4,
    Routev6,
    Service,
    ServiceRoute,
    Static,
    StaticNatDirection,
    SubnetStatic,
    SvcType,
)
from catalystwan.utils.device_model import DeviceModel

basic_cisco_vpn = CiscoVPNModel(
    template_name="Basic_Cisco_VPN_Model", template_description="Primitive", device_models=[DeviceModel.VEDGE_C8000V]
)  # type: ignore


complex_vpn_model = CiscoVPNModel(
    template_name="complex_cisco_vpn",
    template_description="NA",
    device_models=[DeviceModel.VEDGE],
    vpn_name="test_vpn_name",
    omp_admin_distance_ipv4=10,
    omp_admin_distance_ipv6=100,
    route_v4=[Routev4(prefix="prefixv4", next_hop=[NextHop(address="1.1.1.1")])],
    route_v6=[Routev6(prefix="prefixv6", next_hop=[NextHopv6(address="2.2.2.2")], nat=Nat.NAT64)],
    dns=[Dns(dns_addr="1.1.1.1"), Dns(dns_addr="2.2.2.2", role=Role.SECONDARY)],
    dns_ipv6=[DnsIpv6(dns_addr="30a8:b25e:3db5:fe9f:231f:7478:4181:9234")],
    host=[Host(hostname="test_hostname", ip=["1.1.1.1"])],
    service=[
        Service(
            svc_type=SvcType.APPQOE,
            address=["1.1.1.1"],
            interface="Gig0/0/1",
            track_enable=False,
        ),
        Service(
            svc_type=SvcType.FW,
            address=["1.1.122.1", "2.2.2.2"],
            interface="Gig0/0/2",
            track_enable=True,
        ),
        Service(
            svc_type=SvcType.IDP,
            address=["1.1.122.2", "3.2.2.2"],
            interface="Gig0/0/3",
            track_enable=False,
        ),
    ],
    service_route=[
        ServiceRoute(prefix="service_route", vpn=1),
        ServiceRoute(prefix="service_route100", vpn=100),
    ],
    gre_route=[
        GreRoute(prefix="gre_route", vpn=100),
        GreRoute(prefix="gre_route2", vpn=2, interface=["Gig0/0/1", "ge0/0"]),
    ],
    ipsec_route=[
        IpsecRoute(prefix="ipsec-prefix", vpn=10, interface=["ge0/0", "Gig0/0/1"]),
        IpsecRoute(prefix="prefix-2", vpn=100),
    ],
    advertise=[
        Advertise(
            protocol=AdvertiseProtocol.AGGREGATE,
            route_policy="route-policy",
            protocol_sub_type=[AdvertiseProtocolSubType.EXTERNAL],
            prefix_list=[
                PrefixList(
                    prefix_entry="prefix_entry",
                    aggregate_only=True,
                    region=Region.ACCESS,
                )
            ],
        )
    ],
    ipv6_advertise=[
        Ipv6Advertise(
            protocol=Ipv6AdvertiseProtocol.AGGREGATE,
            route_policy="route-policyv6",
            protocol_sub_type=[Ipv6AdvertiseProtocolSubType.EXTERNAL],
            prefix_list=[
                PrefixList(
                    prefix_entry="prefix_entryv6",
                    aggregate_only=False,
                    region=Region.CORE,
                )
            ],
        ),
        Ipv6Advertise(
            protocol=Ipv6AdvertiseProtocol.CONNECTED,
            route_policy="route-policyv6-connected",
            protocol_sub_type=[Ipv6AdvertiseProtocolSubType.EXTERNAL],
            prefix_list=[
                PrefixList(
                    prefix_entry="prefix_entryv6-connected",
                    aggregate_only=True,
                    region=Region.ACCESS,
                )
            ],
        ),
    ],
    pool=[
        Pool(
            name="pool",
            start_address="1.1.1.1",
            end_address="10.10.10.10",
            overload=False,
            leak_from_global=True,
            leak_from_global_protocol=LeakFromGlobalProtocol.CONNECTED,
            leak_to_global=False,
        )
    ],
    natpool=[
        Natpool(
            name=1,
            prefix_length=24,
            range_start="10",
            range_end="100",
            overload=Overload.FALSE,
            direction=Direction.INSIDE,
            tracker_id=10,
        ),
        Natpool(
            name=2,
            prefix_length=24,
            range_start="10",
            range_end="100",
            overload=Overload.TRUE,
            direction=Direction.OUTSIDE,
        ),
    ],
    static=[
        Static(
            pool_name=1,
            source_ip="1.1.1.1",
            translate_ip="1.1.1.2",
            static_nat_direction=StaticNatDirection.INSIDE,
            tracker_id=1,
        ),
        Static(
            pool_name=2,
            source_ip="2.1.1.1",
            translate_ip="2.1.1.2",
            static_nat_direction=StaticNatDirection.OUTSIDE,
        ),
    ],
    subnet_static=[
        SubnetStatic(
            source_ip_subnet="1.1.1.1",
            translate_ip_subnet="2.2.2.2",
            prefix_length=24,
            static_nat_direction=StaticNatDirection.OUTSIDE,
        ),
        SubnetStatic(
            source_ip_subnet="1.1.2.1",
            translate_ip_subnet="2.3.2.2",
            prefix_length=24,
            static_nat_direction=StaticNatDirection.INSIDE,
            tracker_id=10,
        ),
    ],
    port_forward=[
        PortForward(
            pool_name=1,
            source_port=1000,
            translate_port=2000,
            source_ip="1.1.1.1",
            translate_ip="2.2.2.2",
            proto=Proto.TCP,
        ),
        PortForward(
            pool_name=2,
            source_port=1000,
            translate_port=2000,
            source_ip="1.1.4.1",
            translate_ip="2.2.3.2",
            proto=Proto.UDP,
        ),
    ],
    route_import=[
        RouteImport(
            protocol=RouteImportProtocol.BGP,
            protocol_sub_type=[RouteImportProtocolSubType.EXTERNAL],
            route_policy="test_route_policy",
            redistribute=[
                RouteImportRedistribute(
                    protocol=RouteImportRedistributeProtocol.EIGRP,
                    route_policy="test_route_policy",
                )
            ],
        )
    ],
    route_import_from=[
        RouteImportFrom(
            source_vpn=1,
            protocol=RouteImportFromProtocol.CONNECTED,
            protocol_sub_type=[RouteImportFromProtocolSubType.EXTERNAL],
            route_policy="test_route_policy",
            redistribute=[RouteImportFromRedistribute(protocol=RouteImportFromRedistributeProtocol.BGP)],
        ),
        RouteImportFrom(
            source_vpn=100,
            protocol=RouteImportFromProtocol.BGP,
            protocol_sub_type=[RouteImportFromProtocolSubType.EXTERNAL],
            route_policy="test_route_policy",
            redistribute=[
                RouteImportFromRedistribute(
                    protocol=RouteImportFromRedistributeProtocol.EIGRP,
                    route_policy="test_route_policy",
                )
            ],
        ),
    ],
    route_export=[
        RouteExport(
            protocol=RouteExportProtocol.STATIC,
            protocol_sub_type=[RouteExportProtocolSubType.EXTERNAL],
            redistribute=[
                RouteExportRedistribute(
                    protocol=RouteExportRedistributeProtocol.OSPF,
                    route_policy="test_route_policy",
                )
            ],
        )
    ],
)
